#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum
from typing import Any, Generic, List, Literal, Optional, TypeVar, Union

from pydantic import BaseModel, Field

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated


class QuestionType(str, Enum):
    text = "text"
    rating = "rating"
    label_selection = "label_selection"
    multi_label_selection = "multi_label_selection"
    ranking = "ranking"


class ResponseValue(BaseModel):
    value: Any


class BaseQuestionSettings(BaseModel):
    def check_response(self, response: ResponseValue):
        pass


class TextQuestionSettings(BaseQuestionSettings):
    type: Literal[QuestionType.text]
    use_markdown: bool = False

    def check_response(self, response: ResponseValue):
        if not isinstance(response.value, str):
            raise ValueError(f"Expected text value, found {type(response.value)}")


class RatingQuestionSettingsOption(BaseModel):
    value: int


T = TypeVar("T")


class ValidOptionCheckerMixin(BaseQuestionSettings, Generic[T]):
    @property
    def option_values(self) -> List[T]:
        return [option.value for option in self.options]

    def check_response(self, response: ResponseValue):
        if response.value not in self.option_values:
            raise ValueError(f"{response.value!r} is not a valid option.\nValid options are: {self.option_values!r}")


class RatingQuestionSettings(ValidOptionCheckerMixin[int]):
    type: Literal[QuestionType.rating]
    options: List[RatingQuestionSettingsOption]


class ValueTextQuestionSettingsOption(BaseModel):
    value: str
    text: str
    description: Optional[str] = None


class LabelSelectionQuestionSettings(ValidOptionCheckerMixin[str]):
    type: Literal[QuestionType.label_selection]
    options: List[ValueTextQuestionSettingsOption]
    visible_options: Optional[int] = None


def _are_all_elements_in_list(elements: List[T], list_: List[T]) -> List[T]:
    return sorted(list(set(elements) - set(list_)))


class MultiLabelSelectionQuestionSettings(LabelSelectionQuestionSettings):
    type: Literal[QuestionType.multi_label_selection]

    def check_response(self, response: ResponseValue):
        if not isinstance(response.value, list):
            raise ValueError(
                f"This MultiLabelSelection question expects a list of values, found {type(response.value)}"
            )

        if len(response.value) == 0:
            raise ValueError("This MultiLabelSelection question expects a list of values, found empty list")

        unique_values = set(response.value)
        if len(unique_values) != len(response.value):
            raise ValueError(
                "This MultiLabelSelection question expects a list of unique values, but duplicates were found"
            )

        invalid_options = _are_all_elements_in_list(response.value, self.option_values)
        if invalid_options:
            raise ValueError(
                f"{invalid_options!r} are not valid options for this MultiLabelSelection question.\nValid options are:"
                f" {self.option_values!r}"
            )


class RankingQuestionSettings(ValidOptionCheckerMixin[str]):
    type: Literal[QuestionType.ranking]
    options: List[ValueTextQuestionSettingsOption]

    @property
    def rank_values(self) -> List[int]:
        return list(range(1, len(self.option_values) + 1))

    def check_response(self, response: ResponseValue):
        if not isinstance(response.value, list):
            raise ValueError(f"This Ranking question expects a list of values, found {type(response.value)}")

        if len(response.value) != len(self.option_values):
            raise ValueError(
                f"This Ranking question expects a list containing {len(self.option_values)} values, found a list of"
                f" {len(response.value)} values"
            )

        values = []
        ranks = []
        for response_option in response.value:
            values.append(response_option.get("value"))
            ranks.append(response_option.get("rank"))

        invalid_ranks = _are_all_elements_in_list(ranks, self.rank_values)
        if invalid_ranks:
            raise ValueError(
                f"{invalid_ranks!r} are not valid ranks for this Ranking question.\nValid ranks are:"
                f" {self.rank_values!r}"
            )

        invalid_values = _are_all_elements_in_list(values, self.option_values)
        if invalid_values:
            raise ValueError(
                f"{invalid_values!r} are not valid options for this Ranking question.\nValid options are:"
                f" {self.option_values!r}"
            )

        unique_values = set(values)
        if len(response.value) != len(unique_values):
            raise ValueError("This Ranking question expects a list of unique values, but duplicates were found")


QuestionSettings = Annotated[
    Union[
        TextQuestionSettings,
        RatingQuestionSettings,
        LabelSelectionQuestionSettings,
        MultiLabelSelectionQuestionSettings,
        RankingQuestionSettings,
    ],
    Field(..., discriminator="type"),
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import re
from enum import Enum
from typing import Any, Dict, List, Optional

from luqum.elasticsearch import ElasticsearchQueryBuilder, SchemaAnalyzer
from luqum.parser import parser

from argilla.server.daos.backend.query_helpers import filters
from argilla.server.daos.backend.search.model import (
    BackendDatasetsQuery,
    BackendQuery,
    BackendRecordsQuery,
    BaseDatasetsQuery,
    QueryRange,
    SortableField,
    SortConfig,
)


class HighlightParser:
    _SEARCH_KEYWORDS_FIELD = "search_keywords"

    __HIGHLIGHT_PRE_TAG__ = "<@@-ar-key>"
    __HIGHLIGHT_POST_TAG__ = "</@@-ar-key>"
    __HIGHLIGHT_VALUES_REGEX__ = re.compile(rf"{__HIGHLIGHT_PRE_TAG__}(.+?){__HIGHLIGHT_POST_TAG__}")

    __HIGHLIGHT_PHRASE_PRE_PARSER_REGEX__ = re.compile(rf"{__HIGHLIGHT_POST_TAG__}\s+{__HIGHLIGHT_PRE_TAG__}")

    @property
    def search_keywords_field(self) -> str:
        return self._SEARCH_KEYWORDS_FIELD

    @classmethod
    def build_query_highlight(cls):
        return {
            "pre_tags": [cls.__HIGHLIGHT_PRE_TAG__],
            "post_tags": [cls.__HIGHLIGHT_POST_TAG__],
            "require_field_match": True,
            "fields": {
                "text": {},
                "text.*": {},
                "inputs.*": {},
            },
        }

    @classmethod
    def parse_highligth_results(
        cls,
        doc: Dict[str, Any],
        is_phrase_query: bool = False,
    ) -> Optional[List[str]]:
        highlight_info = doc.get("highlight")
        if not highlight_info:
            return None

        search_keywords = []
        for content in highlight_info.values():
            if not isinstance(content, list):
                content = [content]
            text = " ".join(content)

            if is_phrase_query:
                text = re.sub(
                    pattern=cls.__HIGHLIGHT_PHRASE_PRE_PARSER_REGEX__,
                    repl=" ",
                    string=text,
                )
            search_keywords.extend(
                re.findall(
                    pattern=cls.__HIGHLIGHT_VALUES_REGEX__,
                    string=text,
                )
            )
        return list(set(search_keywords))


class EsQueryBuilder:
    _INSTANCE: "EsQueryBuilder" = None
    _LOGGER = logging.getLogger(__name__)

    @classmethod
    def get_instance(cls):
        if not cls._INSTANCE:
            cls._INSTANCE = cls()
        return cls._INSTANCE

    def _datasets_to_es_query(self, query: Optional[BackendDatasetsQuery] = None) -> Dict[str, Any]:
        if not query:
            return filters.match_all()

        query_filters = []
        if query.workspaces:
            query_filters.append(
                filters.boolean_filter(
                    should_filters=[
                        filters.terms_filter("owner.keyword", query.workspaces),  # backward comp.
                        filters.terms_filter("workspace.keyword", query.workspaces),
                    ]
                )
            )

        if query.tasks:
            query_filters.append(
                filters.terms_filter(
                    field="task.keyword",
                    values=query.tasks,
                )
            )
        if query.name:
            query_filters.append(
                filters.term_filter(
                    field="name.keyword",
                    value=query.name,
                )
            )
        if not query_filters:
            return filters.match_all()

        return filters.boolean_filter(
            should_filters=query_filters,
            minimum_should_match=len(query_filters),
        )

    def _search_to_es_query(
        self,
        schema: Optional[Dict[str, Any]] = None,
        query: Optional[BackendRecordsQuery] = None,
        sort: Optional[SortConfig] = None,
    ):
        if not query:
            return filters.match_all()

        if not query.advanced_query_dsl or not query.query_text:
            return self._to_es_query(query)

        text_search = query.query_text
        new_query = query.copy(update={"query_text": None})

        schema = SchemaAnalyzer(schema)
        es_query_builder = ElasticsearchQueryBuilder(
            **{
                **schema.query_builder_options(),
                "default_field": "text",
            }
        )

        query_tree = parser.parse(text_search)
        query_text = es_query_builder(query_tree)
        boolean_filter_query = self._to_es_query(new_query)
        return filters.boolean_filter(
            filter_query=boolean_filter_query,
            must_query=query_text,
        )

    def map_2_es_query(
        self,
        schema: Dict[str, Any],
        query: BackendQuery,
        sort: SortConfig = SortConfig(),
        exclude_fields: Optional[List[str]] = None,
        include_fields: List[str] = None,
        doc_from: Optional[int] = None,
        highlight: Optional[HighlightParser] = None,
        size: Optional[int] = None,
        search_after_param: Optional[Any] = None,
    ) -> Dict[str, Any]:
        if query and query.raw_query:
            es_query = {"query": query.raw_query}
        else:
            es_query: Dict[str, Any] = (
                {"query": self._datasets_to_es_query(query)}
                if isinstance(query, BaseDatasetsQuery)
                else {"query": self._search_to_es_query(schema, query)}
            )

        if search_after_param:
            es_query["search_after"] = search_after_param

        if sort.shuffle:
            self._setup_random_score(es_query)
        else:
            es_sort = self.map_2_es_sort_configuration(schema=schema, sort=sort)
            if es_sort:
                es_query["sort"] = es_sort

        if doc_from:
            es_query["from"] = doc_from

        source = {}
        if exclude_fields:
            source.update({"excludes": exclude_fields})
        if include_fields:
            source.update({"includes": include_fields})
        if source:
            es_query["_source"] = source

        if highlight and (
            not include_fields
            # Enable highlight when requesting ALL fields
            # or specific request the search_keywords one
            or "*" in include_fields
            or highlight.search_keywords_field in include_fields
        ):
            es_query["highlight"] = highlight.build_query_highlight()

        if hasattr(query, "vector") and query.vector is not None:
            self._build_knn_configuration(
                es_query=es_query,
                vector_field=self.get_vector_field_name(query.vector.name),
                vector_value=query.vector.value,
                top_k=query.vector.k or size,
            )

        return es_query

    def map_2_es_sort_configuration(self, schema: Dict[str, Any], sort: SortConfig) -> Optional[List[Dict[str, Any]]]:
        if not sort.sort_by or sort.shuffle:
            return None

        # TODO(@frascuchon): compute valid list from the schema
        valid_fields = sort.valid_fields or [
            "id",
            "metadata",
            "score",
            "predicted",
            "predicted_as",
            "predicted_by",
            "annotated_as",
            "annotated_by",
            "status",
            "last_updated",
            "event_timestamp",
        ]

        id_field = "id"
        id_keyword_field = "id.keyword"
        schema = schema or {}
        mappings = self._clean_mappings(schema.get("mappings", {}))
        use_id_keyword = "text" == mappings.get("id")

        es_sort = []
        for sortable_field in sort.sort_by or [SortableField(id="id")]:
            if valid_fields:
                if sortable_field.id.split(".")[0] not in valid_fields:
                    raise AssertionError(
                        f"Wrong sort id {sortable_field.id}. Valid values are: " f"{[str(v) for v in valid_fields]}"
                    )
            field = sortable_field.id
            if field == id_field and use_id_keyword:
                field = id_keyword_field
            es_sort.append({field: {"order": sortable_field.order}})

        return es_sort

    @classmethod
    def _to_es_query(cls, query: BackendRecordsQuery) -> Dict[str, Any]:
        if query.ids:
            return filters.ids_filter(query.ids)

        query_text = filters.text_query(query.query_text)
        all_filters = filters.metadata(query.metadata)
        if query.has_annotation:
            all_filters.append(filters.exists_field("annotated_by"))
        if query.has_prediction:
            all_filters.append(filters.exists_field("predicted_by"))

        query_data = query.dict(
            exclude={
                "advanced_query_dsl",
                "vector",
                "query_text",
                "metadata",
                "uncovered_by_rules",
                "has_annotation",
                "has_prediction",
            }
        )
        for key, value in query_data.items():
            if value is None:
                continue
            key_filter = None
            if isinstance(value, dict):
                value = getattr(query, key)  # check the original field type
            if isinstance(value, list):
                key_filter = filters.terms_filter(key, value)
            elif isinstance(value, (str, Enum)):
                key_filter = filters.term_filter(key, value)
            elif isinstance(value, QueryRange):
                key_filter = filters.range_filter(field=key, value_from=value.range_from, value_to=value.range_to)

            else:
                cls._LOGGER.warning(f"Cannot parse query value {value} for key {key}")
            if key_filter:
                all_filters.append(key_filter)

        boolean_filter_query = filters.boolean_filter(
            must_query=query_text or filters.match_all(),
            filter_query=filters.boolean_filter(
                should_filters=all_filters,
                minimum_should_match=len(all_filters),
            )
            if all_filters
            else None,
            must_not_query=filters.boolean_filter(
                should_filters=[filters.text_query(q) for q in query.uncovered_by_rules]
            )
            if hasattr(query, "uncovered_by_rules") and query.uncovered_by_rules
            else None,
        )

        return boolean_filter_query

    def _clean_mappings(self, mappings: Dict[str, Any]):
        if not mappings:
            return {}

        return {
            key: definition.get("type") or self._clean_mappings(definition)
            for key, definition in mappings["properties"].items()
        }

    def _build_knn_configuration(
        self,
        *,
        es_query: Dict[str, Any],
        vector_field: str,
        vector_value: List[float],
        top_k: Optional[int] = None,
    ):
        def compute_num_candidates(k: int):
            if k < 50:
                return 500
            if 50 <= k < 200:
                return 100
            if 200 <= k < 500:
                return 2000
            # > 500
            return 2500

        top_k = top_k or 5
        num_candidates = compute_num_candidates(top_k)

        es_query_filter = es_query["query"]

        es_query["knn"] = {
            "field": vector_field,
            "query_vector": vector_value,
            "k": top_k,  # TODO: parameterize
            "num_candidates": num_candidates,  # TODO: parameterize
            "filter": es_query_filter,
        }
        es_query.pop("sort", None)
        del es_query["query"]

    def _setup_random_score(self, es_query: Dict[str, Any]):
        query = es_query["query"]
        es_query["query"] = {
            "function_score": {
                "query": query,
                "random_score": {},
            }
        }
        return es_query

    def get_vector_field_name(self, vector_name: str) -> str:
        return f"vectors.{vector_name}.value"


class OpenSearchQueryBuilder(EsQueryBuilder):
    def _build_knn_configuration(
        self,
        *,
        es_query: Dict[str, Any],
        vector_field: str,
        vector_value: List[float],
        top_k: Optional[int] = None,
    ):
        top_k = top_k or 5
        knn = {
            vector_field: {
                "vector": vector_value,
                "k": top_k,  # TODO: Input from query
            }
        }
        es_query.pop("sort", None)
        filter = es_query.pop("query")
        es_query.update(
            {
                "query": {"knn": knn},
                "post_filter": filter,
            }
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional

import typer
import yaml
from pydantic import BaseModel, constr
from sqlalchemy.orm import Session

from argilla.server.database import SessionLocal
from argilla.server.models import User, UserRole, Workspace
from argilla.server.security.auth_provider.local.settings import settings
from argilla.server.security.model import USER_USERNAME_REGEX, WORKSPACE_NAME_REGEX


class WorkspaceCreate(BaseModel):
    name: constr(regex=WORKSPACE_NAME_REGEX, min_length=1)


class UserCreate(BaseModel):
    first_name: constr(strip_whitespace=True)
    username: constr(regex=USER_USERNAME_REGEX, min_length=1)
    role: UserRole
    api_key: constr(min_length=1)
    password_hash: constr(min_length=1)
    workspaces: Optional[List[WorkspaceCreate]]


class UsersMigrator:
    def __init__(self, users_filename: str):
        self._users_filename = users_filename

        with open(users_filename) as users_file:
            self._users = yaml.safe_load(users_file.read())

    def migrate(self):
        typer.echo(f"Starting users migration process using file {self._users_filename!r}")

        with SessionLocal() as session, session.begin():
            for user in self._users:
                self._migrate_user(session, user)

            typer.echo("Users migration process successfully finished")

    def _migrate_user(self, session: Session, user: dict):
        typer.echo(f"Migrating User with username {user['username']!r}")

        user_create = self._build_user_create(user)

        session.add(self._build_user(session, user_create))
        session.flush()

    def _build_user_create(self, user: dict):
        return UserCreate(
            first_name=user.get("full_name", ""),
            username=user["username"],
            role=self._user_role(user),
            api_key=user["api_key"],
            password_hash=user["hashed_password"],
            workspaces=[WorkspaceCreate(name=workspace_name) for workspace_name in self._user_workspace_names(user)],
        )

    def _build_user(self, session: Session, user_create: UserCreate):
        return User(
            first_name=user_create.first_name,
            username=user_create.username,
            role=user_create.role,
            api_key=user_create.api_key,
            password_hash=user_create.password_hash,
            workspaces=[self._get_or_new_workspace(session, workspace.name) for workspace in user_create.workspaces],
        )

    def _user_role(self, user: dict):
        if user.get("workspaces") is None:
            return UserRole.owner
        else:
            return UserRole.annotator

    def _user_workspace_names(self, user: dict):
        workspace_names = [workspace_name for workspace_name in user.get("workspaces", [])]

        if user["username"] in workspace_names:
            return workspace_names
        else:
            return [user["username"]] + workspace_names

    def _get_or_new_workspace(self, session: Session, workspace_name: str):
        return session.query(Workspace).filter_by(name=workspace_name).first() or Workspace(name=workspace_name)


def migrate():
    """Migrate users defined in YAML file to database."""
    UsersMigrator(settings.users_db_file).migrate()


if __name__ == "__main__":
    typer.run(migrate())



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional

import typer
from pydantic import constr
from sqlalchemy.orm import Session
from typer import Typer

from argilla.server.contexts import accounts
from argilla.server.database import SessionLocal
from argilla.server.models import User, UserRole, Workspace
from argilla.server.security.model import (
    USER_PASSWORD_MIN_LENGTH,
    UserCreate,
    WorkspaceCreate,
)

USER_API_KEY_MIN_LENGTH = 8


class UserCreateForTask(UserCreate):
    api_key: Optional[constr(min_length=USER_API_KEY_MIN_LENGTH)]
    workspaces: Optional[List[WorkspaceCreate]]


def _get_or_new_workspace(session: Session, workspace_name: str):
    return session.query(Workspace).filter_by(name=workspace_name).first() or Workspace(name=workspace_name)


def role_callback(value: str):
    try:
        return UserRole(value).value
    except ValueError:
        raise typer.BadParameter("Only Camila is allowed")


def password_callback(password: str = None):
    # if password is None:
    #     raise typer.BadParameter("Password must be specified.")
    # if len(password)<USER_PASSWORD_MIN_LENGTH:
    #     raise typer.BadParameter(f"Password must be at least {USER_PASSWORD_MIN_LENGTH} characters long.")
    return password


def api_key_callback(api_key: str):
    # if api_key and len(api_key)<USER_PASSWORD_MIN_LENGTH:
    #     raise typer.BadParameter(f"API key must a string with a minimum length of {USER_API_KEY_MIN_LENGTH} characters.")
    return api_key


def create(
    first_name: str = typer.Option(default=None, help="First name as a string."),
    username: str = typer.Option(
        default=None,
        prompt=True,
        help="Username as a lowercase string without spaces allowing letters, numbers, dashes and underscores.",
    ),
    role: UserRole = typer.Option(
        prompt=True,
        default=UserRole.annotator.value,
        show_default=True,
        help="Role for the user.",
    ),
    password: str = typer.Option(
        default=None,
        prompt=True,
        callback=password_callback,
        help=f"Password as a string with a minimum length of {USER_PASSWORD_MIN_LENGTH} characters.",
    ),
    last_name: str = typer.Option(default=None, help="Last name as a string."),
    api_key: Optional[str] = typer.Option(
        default=None,
        callback=api_key_callback,
        help=f"API key as a string with a minimum length of {USER_API_KEY_MIN_LENGTH} characters. If not specified a secure random API key will be generated",
    ),
    workspace: List[str] = typer.Option(
        default=[], help="A workspace that the user will be a member of (can be used multiple times)."
    ),
):
    """Creates a new user in the Argilla database with provided parameters"""
    with SessionLocal() as session:
        if accounts.get_user_by_username(session, username):
            typer.echo(f"User with username {username!r} already exists in database. Skipping...")
            return

        if accounts.get_user_by_api_key(session, api_key):
            typer.echo(f"User with api_key {api_key!r} already exists in database. Skipping...")
            return

        user_create = UserCreateForTask(
            first_name=first_name,
            last_name=last_name,
            username=username,
            role=role,
            password=password,
            api_key=api_key,
            workspaces=[WorkspaceCreate(name=workspace_name) for workspace_name in workspace],
        )

        user = User(
            first_name=user_create.first_name,
            last_name=user_create.last_name,
            username=user_create.username,
            role=user_create.role,
            password_hash=accounts.hash_password(user_create.password),
            api_key=user_create.api_key,
            workspaces=[_get_or_new_workspace(session, workspace.name) for workspace in user_create.workspaces],
        )

        session.add(user)
        session.commit()
        session.refresh(user)

        typer.echo("User successfully created:")
        typer.echo(f"• first_name: {user.first_name!r}")
        if user.last_name:
            typer.echo(f"• last_name: {user.last_name!r}")
        typer.echo(f"• username: {user.username!r}")
        typer.echo(f"• role: {user.role.value!r}")
        typer.echo(f"• api_key: {user.api_key!r}")
        typer.echo(f"• workspaces: {[workspace.name for workspace in user.workspaces]!r}")


if __name__ == "__main__":
    app = Typer(add_completion=False)
    app.command(no_args_is_help=True)(create)
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import typer
import uvicorn

app = typer.Typer(invoke_without_command=True)


# using callback to ensure it is used as sole command
@app.callback(help="Starts the Argilla FastAPI server.", invoke_without_command=True)
def server(port: int = 6900, host: str = "0.0.0.0", access_log: bool = True):
    uvicorn.run(
        "argilla:app",
        port=port,
        host=host,
        access_log=access_log,
    )


if __name__ == "__main__":
    app()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import logging
import warnings
from asyncio import Future
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from argilla.client.client import Argilla
from argilla.client.datasets import Dataset
from argilla.client.models import (  # TODO Remove TextGenerationRecord
    BulkResponse,
    Record,
)
from argilla.client.sdk.commons import errors
from argilla.client.sdk.v1.datasets.api import list_datasets as list_datasets_api_v1
from argilla.client.sdk.workspaces.api import list_workspaces as list_workspaces_api_v0

_LOGGER = logging.getLogger(__name__)

Api = Argilla  # Backward compatibility


class ArgillaSingleton:
    """The active argilla singleton instance"""

    _INSTANCE: Optional[Argilla] = None

    @classmethod
    def get(cls) -> Argilla:
        if cls._INSTANCE is None:
            return cls.init()
        return cls._INSTANCE

    @classmethod
    def clear(cls):
        cls._INSTANCE = None

    @classmethod
    def init(
        cls,
        api_url: Optional[str] = None,
        api_key: Optional[str] = None,
        workspace: Optional[str] = None,
        timeout: int = 60,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Argilla:
        cls._INSTANCE = None

        cls._INSTANCE = Argilla(
            api_url=api_url,
            api_key=api_key,
            timeout=timeout,
            workspace=workspace,
            extra_headers=extra_headers,
        )

        return cls._INSTANCE


def init(
    api_url: Optional[str] = None,
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    timeout: int = 60,
    extra_headers: Optional[Dict[str, str]] = None,
):
    """Init the Python client.

    We will automatically init a default client for you when calling other client methods.
    The arguments provided here will overwrite your corresponding environment variables.

    Args:
        api_url: Address of the REST API. If `None` (default) and the env variable ``ARGILLA_API_URL`` is not set,
            it will default to `http://localhost:6900`.
        api_key: Authentification key for the REST API. If `None` (default) and the env variable ``ARGILLA_API_KEY``
            is not set, it will default to `argilla.apikey`.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        timeout: Wait `timeout` seconds for the connection to timeout. Default: 60.
        extra_headers: Extra HTTP headers sent to the server. You can use this to customize
            the headers of argilla client requests, like additional security restrictions. Default: `None`.

    Examples:
        >>> import argilla as rg
        >>>
        >>> rg.init(api_url="http://localhost:9090", api_key="4AkeAPIk3Y")
        >>> # Customizing request headers
        >>> headers = {"X-Client-id":"id","X-Secret":"secret"}
        >>> rg.init(api_url="http://localhost:9090", api_key="4AkeAPIk3Y", extra_headers=headers)
    """

    ArgillaSingleton.init(
        api_url=api_url,
        api_key=api_key,
        workspace=workspace,
        timeout=timeout,
        extra_headers=extra_headers,
    )


def log(
    records: Union[Record, Iterable[Record], Dataset],
    name: str,
    workspace: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    batch_size: int = 100,
    verbose: bool = True,
    background: bool = False,
    chunk_size: Optional[int] = None,
    num_threads: int = 0,
    max_retries: int = 3,
) -> Union[BulkResponse, Future]:
    """Logs Records to argilla.

    The logging happens asynchronously in a background thread.

    Args:
        records: The record, an iterable of records, or a dataset to log.
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        tags: A dictionary of tags related to the dataset.
        metadata: A dictionary of extra info for the dataset.
        batch_size: The batch size for a data bulk.
        verbose: If True, shows a progress bar and prints out a quick summary at the end.
        background: If True, we will NOT wait for the logging process to finish and return an ``asyncio.Future``
            object. You probably want to set ``verbose`` to False in that case.
        chunk_size: DEPRECATED! Use `batch_size` instead.
        num_threads: If > 0, will use num_thread separate number threads to batches, sending data concurrently.
                Default to `0`, which means no threading at all.
        max_retries: Number of retries when logging a batch of records if a `httpx.TransportError` occurs.
                Default `3`.

    Returns:
        Summary of the response from the REST API.
        If the ``background`` argument is set to True, an ``asyncio.Future`` will be returned instead.

    Examples:
        >>> import argilla as rg
        >>> record = rg.TextClassificationRecord(
        ...     text="my first argilla example",
        ...     prediction=[('spam', 0.8), ('ham', 0.2)]
        ... )
        >>> rg.log(record, name="example-dataset")
        1 records logged to http://localhost:6900/datasets/argilla/example-dataset
        BulkResponse(dataset='example-dataset', processed=1, failed=0)
        >>>
        >>> # Logging records in the background
        >>> rg.log(record, name="example-dataset", background=True, verbose=False)
        <Future at 0x7f675a1fffa0 state=pending>
    """
    return ArgillaSingleton.get().log(
        records=records,
        name=name,
        workspace=workspace,
        tags=tags,
        metadata=metadata,
        batch_size=batch_size,
        verbose=verbose,
        background=background,
        chunk_size=chunk_size,
        num_threads=num_threads,
        max_retries=max_retries,
    )


async def log_async(
    records: Union[Record, Iterable[Record], Dataset],
    name: str,
    workspace: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    batch_size: int = 100,
    verbose: bool = True,
    chunk_size: Optional[int] = None,
) -> BulkResponse:
    """Logs Records to argilla with asyncio.

    Args:
        records: The record, an iterable of records, or a dataset to log.
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        tags: A dictionary of tags related to the dataset.
        metadata: A dictionary of extra info for the dataset.
        batch_size: The batch size for a data bulk.
        verbose: If True, shows a progress bar and prints out a quick summary at the end.
        chunk_size: DEPRECATED! Use `batch_size` instead.

    Returns:
        Summary of the response from the REST API

    Examples:
        >>> # Log asynchronously from your notebook
        >>> import asyncio
        >>> import argilla as rg
        >>> from argilla.utils import setup_loop_in_thread
        >>> loop, _ = setup_loop_in_thread()
        >>> future_response = asyncio.run_coroutine_threadsafe(
        ...     rg.log_async(my_records, dataset_name), loop
        ... )
    """

    warnings.warn(
        (
            "`log_async` is deprecated and will be removed in next release. "
            "Please, use `log` with `background=True` instead"
        ),
        DeprecationWarning,
    )

    future = ArgillaSingleton.get().log(
        records=records,
        name=name,
        workspace=workspace,
        tags=tags,
        metadata=metadata,
        batch_size=batch_size,
        verbose=verbose,
        chunk_size=chunk_size,
        background=True,
    )

    return await asyncio.wrap_future(future)


def _check_if_feedback_dataset_exists(client: Argilla, name: str, workspace: str) -> bool:
    response = list_workspaces_api_v0(client.http_client.httpx)
    workspace_id = None
    for ws in response.parsed:
        if ws.name == workspace:
            workspace_id = ws.id
            break

    response = list_datasets_api_v1(client.http_client.httpx)
    for dataset in response.parsed:
        if dataset.name == name and dataset.workspace_id == workspace_id:
            return True

    return False


def load(
    name: str,
    workspace: Optional[str] = None,
    query: Optional[str] = None,
    vector: Optional[Tuple[str, List[float]]] = None,
    ids: Optional[List[Union[str, int]]] = None,
    limit: Optional[int] = None,
    sort: Optional[List[Tuple[str, str]]] = None,
    id_from: Optional[str] = None,
    batch_size: int = 250,
    include_vectors: bool = True,
    include_metrics: bool = True,
    as_pandas=None,
) -> Dataset:
    """Loads a argilla dataset.

    Args:
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        query: An ElasticSearch query with the `query string
            syntax <https://argilla.readthedocs.io/en/stable/guides/queries.html>`_
        vector: Vector configuration for a semantic search
        ids: If provided, load dataset records with given ids.
        limit: The number of records to retrieve.
        sort: The fields on which to sort [(<field_name>, 'asc|decs')].
        id_from: If provided, starts gathering the records starting from that Record.
            As the Records returned with the load method are sorted by ID, ´id_from´
            can be used to load using batches.
        batch_size: If provided, load `batch_size` samples per request. A lower batch
            size may help avoid timeouts.
        include_vectors: When set to `False`, indicates that records will be retrieved excluding their vectors,
            if any. By default, this parameter is set to `True`, meaning that vectors will be included.
        include_metrics: When set to `False`, indicates that records will be retrieved excluding their metrics.
            By default, this parameter is set to `True`, meaning that metrics will be included.
        as_pandas: DEPRECATED! To get a pandas DataFrame do
            ``rg.load('my_dataset').to_pandas()``.


    Returns:
        A argilla dataset.

    Examples:
        **Basic Loading: load the samples sorted by their ID**

        >>> import argilla as rg
        >>> dataset = rg.load(name="example-dataset")

        **Iterate over a large dataset:**
            When dealing with a large dataset you might want to load it in batches to optimize memory consumption
            and avoid network timeouts. To that end, a simple batch-iteration over the whole database can be done
            employing the `from_id` parameter. This parameter will act as a delimiter, retrieving the N items after
            the given id, where N is determined by the `limit` parameter. **NOTE** If
            no `limit` is given the whole dataset after that ID will be retrieved.

        >>> import argilla as rg
        >>> dataset_batch_1 = rg.load(name="example-dataset", limit=1000)
        >>> dataset_batch_2 = rg.load(name="example-dataset", limit=1000, id_from=dataset_batch_1[-1].id)

    """
    argilla = ArgillaSingleton.get()
    try:
        return argilla.load(
            name=name,
            workspace=workspace,
            query=query,
            vector=vector,
            ids=ids,
            limit=limit,
            sort=sort,
            id_from=id_from,
            batch_size=batch_size,
            include_metrics=include_metrics,
            include_vectors=include_vectors,
            as_pandas=as_pandas,
        )
    except errors.NotFoundApiError as e:
        workspace = workspace or argilla.get_workspace()
        if _check_if_feedback_dataset_exists(client=argilla, name=name, workspace=workspace):
            raise ValueError(
                f"The dataset '{name}' exists but it is a `FeedbackDataset`. Use `rg.FeedbackDataset.from_argilla`"
                " instead to load it."
            )
        raise e


def copy(
    dataset: str,
    name_of_copy: str,
    workspace: str = None,
):
    """
    Creates a copy of a dataset including its tags and metadata

    Args:
        dataset: Name of the source dataset
        name_of_copy: Name of the copied dataset
        workspace: If provided, dataset will be copied to that workspace

    Examples:
        >>> import argilla as rg
        >>> rg.copy("my_dataset", name_of_copy="new_dataset")
        >>> rg.load("new_dataset")
    """
    return ArgillaSingleton.get().copy(
        dataset=dataset,
        name_of_copy=name_of_copy,
        workspace=workspace,
    )


def delete(name: str, workspace: Optional[str] = None):
    """
    Deletes a dataset.

    Args:
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.

    Examples:
        >>> import argilla as rg
        >>> rg.delete(name="example-dataset")
    """
    ArgillaSingleton.get().delete(name=name, workspace=workspace)


def delete_records(
    name: str,
    workspace: Optional[str] = None,
    query: Optional[str] = None,
    ids: Optional[List[Union[str, int]]] = None,
    discard_only: bool = False,
    discard_when_forbidden: bool = True,
) -> Tuple[int, int]:
    """Delete records from a argilla dataset.

    Args:
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        query: An ElasticSearch query with the `query string syntax
            <https://docs.argilla.io/en/latest/guides/query_datasets.html>`_
        ids: If provided, deletes dataset records with given ids.
        discard_only: If `True`, matched records won't be deleted. Instead, they will be marked as `Discarded`
        discard_when_forbidden: Only super-user or dataset creator can delete records from a dataset.
            So, running "hard" deletion for other users will raise an `ForbiddenApiError` error.
            If this parameter is `True`, the client API will automatically try to mark as ``Discarded``
            records instead. Default, `True`

    Returns:
        The total of matched records and real number of processed errors. These numbers could not
        be the same if some data conflicts are found during operations (some matched records change during
        deletion).

    Examples:
        >>> ## Delete by id
        >>> import argilla as rg
        >>> rg.delete_records(name="example-dataset", ids=[1,3,5])
        >>> ## Discard records by query
        >>> import argilla as rg
        >>> rg.delete_records(name="example-dataset", query="metadata.code=33", discard_only=True)
    """
    return ArgillaSingleton.get().delete_records(
        name=name,
        workspace=workspace,
        query=query,
        ids=ids,
        discard_only=discard_only,
        discard_when_forbidden=discard_when_forbidden,
    )


def set_workspace(workspace: str):
    """Sets the active workspace.

    Args:
        workspace: The new workspace
    """
    ArgillaSingleton.get().set_workspace(workspace=workspace)


def get_workspace() -> str:
    """Returns the name of the active workspace.

    Returns:
        The name of the active workspace as a string.
    """
    return ArgillaSingleton.get().get_workspace()


def active_client() -> Argilla:
    """Returns the active argilla client.

    If Active client is None, initialize a default one.
    """
    return ArgillaSingleton.get()


active_api = active_client  # backward compatibility



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import random
from abc import abstractmethod
from collections import Counter
from enum import Enum
from typing import Any, Dict, List, Union

from pydantic import BaseModel, root_validator, validator

from argilla.client.feedback.schemas import (
    FeedbackRecord,
    LabelQuestion,
    MultiLabelQuestion,
    RatingQuestion,
    ValueSchema,
)


class UnifiedValueSchema(ValueSchema):
    """A value schema for a unification value.

    Args:
        value (Union[StrictStr, StrictInt, List[str]]): The unification value of the record.
        strategy (Literal["mean", "majority", "max", "min"]): The strategy to unify the responses. Defaults to "majority".

    Examples:
        >>> import argilla as rg
        >>> value = rg.UnifiedValueSchema(value="Yes", strategy="majority")
        >>> # or use a dict
        >>> value = {"value": "Yes", "strategy": "majority"}
    """

    strategy: Union["RatingQuestionStrategy", "LabelQuestionStrategy", "MultiLabelQuestionStrategy"]


class RatingQuestionStrategy(Enum):
    """
    Options:
        - "mean": the mean value of the ratings
        - "majority": the majority value of the ratings
        - "max": the max value of the ratings
        - "min": the min value of the ratings
    """

    MEAN: str = "mean"
    MAJORITY: str = "majority"
    MAX: str = "max"
    MIN: str = "min"

    def unify_responses(self, records: List[FeedbackRecord], question: RatingQuestion):
        UnifiedValueSchema.update_forward_refs()
        # check if field is a str or a RatingQuestion
        if isinstance(question, str):
            pass
        elif isinstance(question, RatingQuestion):
            question = question.name
        else:
            raise ValueError("Invalid field type. Must be a str or RatingQuestion")
        # choose correct unification method
        if self.value == self.MAJORITY.value:
            return self._majority(records, question)
        else:
            return self._aggregate(records, question)

    def _aggregate(self, records: List[FeedbackRecord], question: str):
        for rec in records:
            if not rec.responses:
                continue
            # only allow for submitted responses
            responses = [resp for resp in rec.responses if resp.status == "submitted"]
            # get responses with a value that is most frequent
            ratings = []
            for resp in responses:
                if question in resp.values:
                    ratings.append(int(resp.values[question].value))
            if not ratings:
                continue
            # unified response
            if self.value == self.MEAN.value:
                unified_value = str(int(sum(ratings) / len(ratings)))
            elif self.value == self.MAX.value:
                unified_value = str(max(ratings))
            elif self.value == self.MIN.value:
                unified_value = str(min(ratings))
            else:
                raise ValueError("Invalid aggregation method")
            rec._unified_responses[question] = [UnifiedValueSchema(value=unified_value, strategy=self.value)]
        return records

    def _majority(self, records: List[FeedbackRecord], question: str):
        for rec in records:
            if not rec.responses:
                continue
            counter = Counter()
            # only allow for submitted responses
            responses = [resp for resp in rec.responses if resp.status == "submitted"]
            # get responses with a value that is most frequent
            for resp in responses:
                if question in resp.values:
                    counter.update([resp.values[question].value])
            if not counter.values():
                continue
            # Find the maximum count
            max_count = max(counter.values())
            # Get a list of values with the maximum count
            most_common_values = [value for value, count in counter.items() if count == max_count]
            if len(most_common_values) > 1:
                majority_value = random.choice(most_common_values)
            else:
                majority_value = counter.most_common(1)[0][0]
            rec._unified_responses[question] = [UnifiedValueSchema(value=majority_value, strategy=self.value)]
        return records


class LabelQuestionStrategyMixin:
    def unify_responses(self, records: List[FeedbackRecord], question: Union[str, LabelQuestion, MultiLabelQuestion]):
        UnifiedValueSchema.update_forward_refs()
        # check if field is a str or a LabelQuestion
        if isinstance(question, (LabelQuestion, MultiLabelQuestion)):
            question = question.name
        elif isinstance(question, str):
            pass
        else:
            raise ValueError("Invalid field type. Must be a str, LabelQuestion, MultiLabelQuestion")
        # choose correct unification method
        if self.value == self.MAJORITY.value:
            return self._majority(records, question)
        elif self.value == self.MAJORITY_WEIGHTED.value:
            return self._majority_weighted(records, question)
        elif self.value == self.DISAGREEMENT.value:
            return self._disagreement(records, question)

    @abstractmethod
    def _majority(self, records: List[FeedbackRecord], question: str):
        """Must be implemented by subclasses"""

    @abstractmethod
    def _majority_weighted(self, records: List[FeedbackRecord], question: str):
        """Must be implemented by subclasses"""

    def _disagreement(self, records: List[FeedbackRecord], question: str):
        unified_records = []
        for rec in records:
            if not rec.responses:
                continue
            # only allow for submitted responses
            responses = [resp for resp in rec.responses if resp.status == "submitted"]
            # get responses with a value that is most frequent
            if question not in rec._unified_responses:
                rec._unified_responses[question] = []
            for resp in responses:
                if question not in resp.values:
                    continue
                else:
                    rec._unified_responses[question].append(
                        UnifiedValueSchema(value=resp.values[question].value, strategy=self.value)
                    )
        return unified_records


class LabelQuestionStrategy(LabelQuestionStrategyMixin, Enum):
    """
    Options:
        - "majority": the majority value of the labels
        - "majority_weighted": the majority value of the labels, weighted by annotator's confidence
        - "disagreement": preserve the natural disagreement between annotators

    Examples:
        >>> from argilla import LabelQuestion, LabelQuestionStrategy
        >>> strategy = LabelQuestionStrategy("majority")
        >>> records = strategy.unify_responses(records, question=LabelQuestion(...))
    """

    MAJORITY: str = "majority"
    MAJORITY_WEIGHTED: str = "majority_weighted"
    DISAGREEMENT: str = "disagreement"

    def _majority(self, records: List[FeedbackRecord], question: str):
        for rec in records:
            if not rec.responses:
                continue
            counter = Counter()
            # only allow for submitted responses
            responses = [resp for resp in rec.responses if resp.status == "submitted"]
            # get responses with a value that is most frequent
            for resp in responses:
                if question in resp.values:
                    counter.update([resp.values[question].value])
            if not counter.values():
                continue
            # Find the maximum count
            max_count = max(counter.values())
            # Get a list of values with the maximum count
            most_common_values = [value for value, count in counter.items() if count == max_count]
            if len(most_common_values) > 1:
                majority_value = random.choice(most_common_values)
            else:
                majority_value = counter.most_common(1)[0][0]

            rec._unified_responses[question] = [UnifiedValueSchema(value=majority_value, strategy=self.value)]
        return rec

    @classmethod
    def _majority_weighted(self, records: List[FeedbackRecord], question: LabelQuestion):
        raise NotImplementedError("'majority_weighted'-strategy not implemented yet")


class MultiLabelQuestionStrategy(LabelQuestionStrategyMixin, Enum):
    """
    Options:
        - "majority": the majority value of the labels
        - "majority_weighted": the majority value of the labels, weighted by annotator's confidence
        - "disagreement": preserve the natural disagreement between annotators
    """

    MAJORITY: str = "majority"
    MAJORITY_WEIGHTED: str = "majority_weighted"
    DISAGREEMENT: str = "disagreement"

    def _majority(self, records: List[FeedbackRecord], question: str):
        for rec in records:
            if not rec.responses:
                continue
            counter = Counter()
            # only allow for submitted responses
            responses = [resp for resp in rec.responses if resp.status == "submitted"]
            # get responses with a value that is most frequent
            for resp in responses:
                if isinstance(resp.values[question].value, list):
                    for value in resp.values[question].value:
                        counter.update([value])
                else:
                    counter.update([resp.values[question].value])
            # check if there is a majority based on the number of responses
            majority = int(len(responses) // 2) + 1
            majority_value = []
            for value, count in counter.items():
                if count >= majority:
                    majority_value.append(value)

            rec._unified_responses[question] = [UnifiedValueSchema(value=majority_value, strategy=self.value)]
        return records

    @classmethod
    def _majority_weighted(self, records: List[FeedbackRecord], question: MultiLabelQuestion):
        raise NotImplementedError("'majority_weighted'-strategy not implemented yet")


class RatingQuestionUnification(BaseModel):
    """Rating unification for a rating question

    Args:
        question (RatingQuestion): rating question
        strategy (Union[str, RatingQuestionStrategy]): unification strategy. Defaults to "mean".
            mean (str): the mean value of the ratings.
            majority (str): the majority value of the ratings.
            max (str): the max value of the ratings
            min (str): the min value of the ratings

    Examples:
        >>> from argilla import RatingQuestion, RatingUnification, RatingQuestionStrategy
        >>> RatingUnification(question=RatingQuestion(...), strategy="mean")
        >>> # or use a RatingQuestionStrategy
        >>> RatingUnification(question=RatingQuestion(...), strategy=RatingQuestionStrategy.MEAN)
    """

    question: RatingQuestion
    strategy: Union[str, RatingQuestionStrategy] = "mean"

    @validator("strategy", always=True)
    def strategy_must_be_valid(cls, v: Union[str, RatingQuestionStrategy]) -> RatingQuestionStrategy:
        if isinstance(v, str):
            return RatingQuestionStrategy(v)
        return v


class LabelQuestionUnification(BaseModel):
    """Label unification for a label question

    Args:
        question (Union[LabelQuestion, MultiLabelQuestion]): label question
        strategy (Union[str, LabelQuestionStrategy, MultiLabelQuestionStrategy]): unification strategy. Defaults to "majority".
            majority (str): the majority value of the labels
            majority_weighted (str): the majority value of the labels, weighted by annotator's confidence
            disagreement (str): preserve the natural disagreement between annotators

    Examples:
        >>> from argilla import LabelQuestion, LabelUnification, LabelQuestionStrategy
        >>> LabelUnification(question=LabelQuestion(...), strategy="majority")
        >>> # or use a LabelQuestionStrategy
        >>> LabelUnification(question=LabelQuestion(...), strategy=LabelQuestionStrategy.MAJORITY)
    """

    question: Union[LabelQuestion, MultiLabelQuestion]
    strategy: Union[str, LabelQuestionStrategy, MultiLabelQuestionStrategy] = "majority"

    def unify_responses(self, records: List[FeedbackRecord]):
        return self.strategy.unify_responses(records, self.question)

    @root_validator
    def strategy_must_be_valid_and_align_with_question(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        strategy = values.get("strategy", "majority")
        question = values.get("question")
        if isinstance(question, LabelQuestion):
            if isinstance(strategy, str):
                strategy = LabelQuestionStrategy(strategy)
            if strategy.value in LabelQuestionStrategy.__members__:
                raise ValueError(f"Only LabelQuestionStrategy is compatible with LabelQuestion not {strategy}")
        elif isinstance(question, MultiLabelQuestion):
            if isinstance(strategy, str):
                strategy = MultiLabelQuestionStrategy(strategy)
            if strategy.value in MultiLabelQuestionStrategy.__members__:
                raise ValueError(
                    f"Only MultiLabelQuestionStrategy is compatible with MultiLabelQuestion not {strategy}"
                )
        values["strategy"] = strategy
        return values


MultiLabelQuestionUnification = LabelQuestionUnification



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import Any, Dict, List, Optional

from argilla.server.daos.backend.metrics.base import (
    ElasticsearchMetric,
    TermsAggregation,
)
from argilla.server.daos.backend.query_helpers import aggregations, filters
from argilla.server.helpers import unflatten_dict


@dataclasses.dataclass
class DatasetLabelingRulesMetric(ElasticsearchMetric):
    def _build_aggregation(self, queries: List[str]) -> Dict[str, Any]:
        rules_filters = [filters.text_query(rule_query) for rule_query in queries]
        return aggregations.filters_aggregation(
            filters={
                "covered_records": filters.boolean_filter(should_filters=rules_filters, minimum_should_match=1),
                "annotated_covered_records": filters.boolean_filter(
                    filter_query=filters.exists_field("annotated_as"),
                    should_filters=rules_filters,
                    minimum_should_match=1,
                ),
            }
        )


@dataclasses.dataclass
class LabelingRulesMetric(ElasticsearchMetric):
    id: str

    def _build_aggregation(self, rule_query: str, labels: Optional[List[str]] = None) -> Dict[str, Any]:
        annotated_records_filter = filters.exists_field("annotated_as")
        rule_query_filter = filters.text_query(rule_query)
        aggr_filters = {
            "covered_records": rule_query_filter,
            "annotated_covered_records": filters.boolean_filter(
                filter_query=annotated_records_filter,
                should_filters=[rule_query_filter],
            ),
        }

        if labels is not None:
            for label in labels:
                rule_label_annotated_filter = filters.term_filter("annotated_as", value=label)
                encoded_label = self._encode_label_name(label)
                aggr_filters.update(
                    {
                        f"{encoded_label}.correct_records": filters.boolean_filter(
                            filter_query=annotated_records_filter,
                            should_filters=[
                                rule_query_filter,
                                rule_label_annotated_filter,
                            ],
                            minimum_should_match=2,
                        ),
                        f"{encoded_label}.incorrect_records": filters.boolean_filter(
                            filter_query=annotated_records_filter,
                            must_query=rule_query_filter,
                            must_not_query=rule_label_annotated_filter,
                        ),
                    }
                )

        return aggregations.filters_aggregation(aggr_filters)

    @staticmethod
    def _encode_label_name(label: str) -> str:
        return label.replace(".", "@@@")

    @staticmethod
    def _decode_label_name(label: str) -> str:
        return label.replace("@@@", ".")

    def aggregation_result(self, aggregation_result: Dict[str, Any]) -> Dict[str, Any]:
        if self.id in aggregation_result:
            aggregation_result = aggregation_result[self.id]

        aggregation_result = unflatten_dict(aggregation_result)
        results = {
            "covered_records": aggregation_result.pop("covered_records"),
            "annotated_covered_records": aggregation_result.pop("annotated_covered_records"),
        }

        all_correct = []
        all_incorrect = []
        all_precision = []
        for label, metrics in aggregation_result.items():
            correct = metrics.get("correct_records", 0)
            incorrect = metrics.get("incorrect_records", 0)
            annotated = correct + incorrect
            metrics["annotated"] = annotated
            if annotated > 0:
                precision = correct / annotated
                metrics["precision"] = precision
                all_precision.append(precision)

            all_correct.append(correct)
            all_incorrect.append(incorrect)
            results[self._decode_label_name(label)] = metrics

        results["correct_records"] = sum(all_correct)
        results["incorrect_records"] = sum(all_incorrect)
        if len(all_precision) > 0:
            results["precision"] = sum(all_precision) / len(all_precision)

        return results


METRICS = {
    "predicted_as": TermsAggregation(
        id="predicted_as",
        field="predicted_as",
    ),
    "annotated_as": TermsAggregation(
        id="annotated_as",
        field="annotated_as",
    ),
    "labeling_rule": LabelingRulesMetric(id="labeling_rule"),
    "dataset_labeling_rules": DatasetLabelingRulesMetric(id="dataset_labeling_rules"),
}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from json import JSONDecodeError
from typing import Any, Dict, Optional, Type, TypeVar, Union

import httpx

from argilla.client.sdk.commons.errors import WrongResponseError
from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import (
    ErrorMessage,
    HTTPValidationError,
    Response,
)


def build_raw_response(
    response: httpx.Response,
) -> Response[Union[Dict[str, Any], ErrorMessage, HTTPValidationError]]:
    return build_typed_response(response)


ResponseType = TypeVar("ResponseType")


def build_typed_response(
    response: httpx.Response,
    response_type_class: Optional[Type[ResponseType]] = None,
) -> Response[Union[ResponseType, ErrorMessage, HTTPValidationError]]:
    parsed_response = check_response(response, expected_response=response_type_class)
    if response_type_class:
        parsed_response = response_type_class(**parsed_response)
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=parsed_response,
    )


def check_response(response: httpx.Response, **kwargs) -> Any:
    if 200 <= response.status_code < 300:
        try:
            return response.json()
        except JSONDecodeError:
            raise WrongResponseError(
                message="Cannot parse json data from response",
                response=response.content,
            )
    if 300 <= response.status_code < 400:
        message = (
            f"Unexpected response {response.status_code} status. "
            "Verify your client/server connection and retry the operation"
        )
        raise WrongResponseError(
            message=message,
            response=response.content,
        )
    handle_response_error(response, **kwargs)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import sys
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel

from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.base import ArgillaTrainerSkeleton
from argilla.utils.dependency import require_version

__all__ = ["ArgillaSpaCyTrainer", "ArgillaSpaCyTransformersTrainer"]


class _ArgillaSpaCyTrainerBase(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaSpaCyTrainer")
    _logger.setLevel(logging.INFO)

    require_version("spacy")

    def __init__(
        self,
        language: Optional[str] = None,
        gpu_id: Optional[int] = -1,
        model: Optional[str] = None,
        optimize: Literal["efficiency", "accuracy"] = "efficiency",
        *args,
        **kwargs,
    ) -> None:
        """Initialize the `_ArgillaSpaCyTrainerBase` class.

        Args:
            dataset: A `spacy.tokens.DocBin` object or a tuple of `spacy.tokens.DocBin` objects.
            record_class:
                A `TextClassificationRecord`, `TokenClassificationRecord`, or `Text2TextRecord`
                object. Defaults to None.
            seed: A `int` with the seed for the random number generator. Defaults to None.
            multi_label: A `bool` indicating whether the task is multi-label or not. Defaults to False.
            language:
                A `str` with the `spaCy` language code e.g. "en". See all the supported languages and their
                codes in `spaCy` at https://spacy.io/usage/models#languages. Defaults to None.
            gpu_id:
                the GPU ID to use. Defaults to -1, which means that the CPU will be used by default.
                GPU IDs start in 0, which stands for the default GPU in the system, if available.
            model:
                A `str` with the `spaCy` model name to use. If it contains vectors it
                can also be used for training/fine-tuning, e.g. "en_core_web_lg"
                contains vectors, while "en_core_web_sm" doesn't. Defaults to None.
            optimize:
                A `str` with the optimization strategy to use. Either "efficiency" or "accuracy".
                Defaults to "efficiency", which means that the model will be smaller, faster,
                and use less memory, but it will be less accurate. If "accuracy" is used, the model
                will be larger, slower, and use more memory, but it will be more accurate.
                Defaults to "efficiency".

        Raises:
            NotImplementedError: If the `record_class` is not supported or if the
                `init_training_args` method has not been implemented.
        """
        super().__init__(*args, **kwargs)
        import spacy

        self._nlp = None
        self._model = model

        self.config = {}

        if self._record_class == TokenClassificationRecord:
            self._column_mapping = {
                "text": "text",
                "token": "tokens",
                "ner_tags": "ner_tags",
            }
            self._pipeline = ["ner"]
        elif self._record_class == TextClassificationRecord:
            if self._multi_label:
                self._column_mapping = {"text": "text", "binarized_label": "label"}
                self._pipeline = ["textcat_multilabel"]
            else:
                self._column_mapping = {"text": "text", "label": "label"}
                self._pipeline = ["textcat"]
        else:
            raise NotImplementedError("`rg.Text2TextRecord` is not supported yet.")

        self._train_dataset, self._eval_dataset = (
            self._dataset if isinstance(self._dataset, tuple) and len(self._dataset) > 1 else (self._dataset, None)
        )
        self._train_dataset_path = "./train.spacy"
        self._eval_dataset_path = "./dev.spacy" if self._eval_dataset else "./train.spacy"

        self.language = language or "en"
        self.optimize = optimize

        self.gpu_id = gpu_id
        self.use_gpu = False
        if self.gpu_id != -1:
            self.use_gpu = spacy.prefer_gpu(self.gpu_id)

        if self.use_gpu:
            try:
                require_version("torch")
                self.has_torch = True
            except Exception:
                self.has_torch = False

            try:
                require_version("tensorflow")
                self.has_tensorflow = True
            except Exception:
                self.has_tensorflow = False

            if not self.has_torch and not self.has_tensorflow:
                self._logger(
                    "Either `torch` or `tensorflow` need to be installed to use the"
                    " GPU, since any of those is required as the GPU allocator. Falling"
                    " back to the CPU."
                )
                self.use_gpu = False
                self.gpu_id = -1

        self.init_training_args()

    def init_model(self):
        import spacy

        self._nlp = spacy.load(self._model)

    def __repr__(self) -> None:
        """Return the string representation of the `ArgillaSpaCyTrainer` object containing
        just the args that can be updated via `update_config`."""
        formatted_string = []
        formatted_string.append(
            "WARNING:`ArgillaSpaCyTrainer.update_config` only supports the update of"
            " the `training` arguments defined in the `config.yaml`."
        )
        formatted_string.append("\n`ArgillaSpaCyTrainer`")
        for key, val in self.config["training"].items():
            if isinstance(val, dict):
                continue
            formatted_string.append(f"\t{key}: {val}")
        return "\n".join(formatted_string)

    def update_config(
        self,
        **spacy_training_config,
    ) -> None:
        """Update the `spaCy` training config.

        Disclaimer: currently just the `training` config is supported, but in the future
        we will support all the `spaCy` config values supported for a more precise control
        over the training process. Also note that the arguments may differ between the CPU
        and GPU training.

        Args:
            **spacy_training_config: The `spaCy` training config.
        """
        self.config["training"].update(spacy_training_config)

    def train(self, output_dir: Optional[str] = None) -> None:
        """Train the pipeline using `spaCy`.

        Args:
            path: A `str` with the path to save the trained pipeline. Defaults to None.
        """
        from spacy.training.initialize import init_nlp
        from spacy.training.loop import train as train_nlp

        self._logger.warning(
            "Note that the spaCy training is expected to be used through the CLI rather"
            " than programatically, so the dataset needs to be dumped into the disk and"
            " then loaded from disk. More information at"
            " https://spacy.io/usage/training#api"
        )
        self._logger.info(f"Dumping the train dataset to {self._train_dataset_path}")
        self._train_dataset.to_disk(self._train_dataset_path)
        if self._eval_dataset:
            self._logger.info(f"Dumping the dev dataset to {self._eval_dataset_path}")
            self._eval_dataset.to_disk(self._eval_dataset_path)

        # Both `init_nlp` and `train_nlp` must be executed in the same Jupyter Notebook
        # cell if using the GPU, otherwise, since `thinc` is using `ContextVars` to
        # store the `Config` object, the `Config` object will be lost between cells and
        # the training will fail.
        self._nlp = init_nlp(self.config, use_gpu=self.gpu_id)
        self._nlp, _ = train_nlp(self._nlp, use_gpu=self.gpu_id, stdout=sys.stdout, stderr=sys.stderr)

        if output_dir:
            self.save(output_dir)

    def save(self, output_dir: str) -> None:
        """Save the trained pipeline to disk.

        Args:
            output_dir: A `str` with the path to save the pipeline.
        """
        output_dir = Path(output_dir) if isinstance(output_dir, str) else output_dir
        if output_dir and not output_dir.exists():
            output_dir.mkdir(parents=True)
        self._nlp.to_disk(output_dir)

    def predict(
        self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs
    ) -> Union[Dict[str, Any], List[Dict[str, Any]], BaseModel, List[BaseModel]]:
        """Predict the labels for the given text using the trained pipeline.

        Args:
            text: A `str` or a `List[str]` with the text to predict the labels for.
            as_argilla_records:
                A `bool` indicating whether to return the predictions as `argilla` records
                or as `dicts`. Defaults to True.

        Returns:
            Either a `dict`, `BaseModel` (if `as_argilla_records` is True) or a `List[dict]`,
            `List[BaseModel]` (if `as_argilla_records` is True) with the predictions.
        """
        if self._nlp is None:
            self._logger.warning("Using model without fine-tuning.")
            self.init_model()

        str_input = False
        if isinstance(text, str):
            text = [text]
            str_input = True

        formatted_prediction = []
        docs = self._nlp.pipe(text, **kwargs)
        if as_argilla_records:
            for doc in docs:
                if "ner" in self._pipeline:
                    entities = [(ent.label_, ent.start_char, ent.end_char) for ent in doc.ents]
                    pred = {
                        "text": doc.text,
                        "tokens": [t.text for t in doc],
                        "prediction": entities,
                    }
                    pred = self._record_class(**pred)
                elif any([p in self._pipeline for p in ["textcat", "textcat_multilabel"]]):
                    pred = {
                        "text": doc.text,
                        "prediction": [(k, v) for k, v in doc.cats.items()],
                    }
                    pred = self._record_class(**pred, multi_label=self._multi_label)
                else:
                    continue
                formatted_prediction.append(pred)
        else:
            formatted_prediction = docs

        if str_input:
            formatted_prediction = list(formatted_prediction)[0]
        return formatted_prediction


class ArgillaSpaCyTrainer(_ArgillaSpaCyTrainerBase):
    def __init__(self, freeze_tok2vec: bool = False, **kwargs) -> None:
        """Initialize the `ArgillaSpaCyTrainer` class.

        Args:
            freeze_tok2vec: A `bool` indicating whether to freeze the `tok2vec` weights
                during the training. Defaults to False.
            **kwargs: The `ArgillaSpaCyTrainerBase` arguments.

        Examples:
            >>> from argilla import ArgillaSpaCyTrainer
            >>> trainer = ArgillaSpaCyTrainer(
        """
        self.freeze_tok2vec = freeze_tok2vec
        super().__init__(**kwargs)

    def init_training_args(self) -> None:
        from spacy.cli.init_config import init_config

        # We generate the config with GPU just when we are using `spacy-transformers`,
        # otherwise the default configuration will be messed up for `spacy`.
        self.config = init_config(
            lang=self.language,
            pipeline=self._pipeline,
            optimize=self.optimize,
            gpu=False,
        )

        self.config["paths"]["train"] = self._train_dataset_path
        self.config["paths"]["dev"] = self._eval_dataset_path
        self.config["system"]["seed"] = self._seed or 42

        # Now we can already set the GPU properties if we want to train/fine-tune a
        # `spacy` model using the GPU, or a `spacy-transformers` model using the CPU.
        self.config["system"]["gpu_allocator"] = (
            ("pytorch" if self.has_torch else "tensorflow" if self.has_tensorflow else None) if self.use_gpu else None
        )
        self.config["nlp"]["batch_size"] = 128 if self.use_gpu else 1000

        if "tok2vec" in self.config["nlp"]["pipeline"]:
            # If we want to fine-tune the `tok2vec` component, then we need to set the
            # `init_tok2vec` path to the model we want to fine-tune.
            if self.freeze_tok2vec is False:
                self.config["paths"]["init_tok2vec"] = self._model
            else:
                # Otherwise, if we don't want to fine-tune the `tok2vec` component, then we
                # need to set the `frozen_components` and `annotating_components` to
                # `["tok2vec"]`.
                self.config["training"]["frozen_components"] = ["tok2vec"]
                self.config["training"]["annotating_components"] = ["tok2vec"]


class ArgillaSpaCyTransformersTrainer(_ArgillaSpaCyTrainerBase):
    require_version("spacy>=3.5.3")  # Required to generate the `spacy-transformers` configuration
    require_version("spacy-transformers")

    def __init__(self, update_transformer: bool = True, **kwargs) -> None:
        """Initialize the `ArgillaSpaCyTransformersTrainer` class.

        Args:
            update_transformer: A `bool` indicating whether to update the transformer
                weights during the training. Defaults to True.
            **kwargs: The `ArgillaSpaCyTrainerBase` arguments.
        """
        self.update_transformer = update_transformer
        super().__init__(**kwargs)

    def init_training_args(self) -> None:
        from spacy.cli.init_config import init_config

        # We generate the config with GPU just when we are using `spacy-transformers`,
        # otherwise the default configuration will be messed up for `spacy`.
        self.config = init_config(
            lang=self.language,
            pipeline=self._pipeline,
            optimize=self.optimize,
            gpu=True,
        )

        self.config["paths"]["train"] = self._train_dataset_path
        self.config["paths"]["dev"] = self._eval_dataset_path
        self.config["system"]["seed"] = self._seed or 42

        # Now we can already set the GPU properties if we want to train/fine-tune a
        # `spacy` model using the GPU, or a `spacy-transformers` model using the CPU.
        self.config["system"]["gpu_allocator"] = (
            ("pytorch" if self.has_torch else "tensorflow" if self.has_tensorflow else None) if self.use_gpu else None
        )
        self.config["nlp"]["batch_size"] = 128 if self.use_gpu else 16

        # If we use `spacy-transformers` then we need to set the `transformer` component
        # in the pipeline, and we need to set the `name` of the model to load.
        self.config["components"]["transformer"]["name"] = self._model
        self.config["nlp"]["pipeline"] = ["transformer"] + self._pipeline

        if "transformer" in self.config["nlp"]["pipeline"]:
            # The `transformer` component cannot be frozen, but we can set the `grad_factor`
            # to 0.0 to avoid updating the weights of the `transformer` component. Even though
            # the computation of those weights will be performed, the gradients will be
            # multiplied by 0.0, so the weights will not be updated.
            # self.config["training"]["frozen_components"] = ["transformer"]
            # self.config["training"]["annotating_components"] = ["transformer"]
            if not self.update_transformer:
                self.config["components"]["transformer"]["grad_factor"] = 0.0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import io
from typing import Optional

import alembic
import typer
from alembic import command
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.util import CommandError

from argilla.server.database import ALEMBIC_CONFIG_FILE, TAGGED_REVISIONS
from argilla.tasks.database import utils


def migrate_db(revision: Optional[str] = typer.Option(default="head", help="DB Revision to migrate to")):
    current_revision = utils.get_current_revision(ALEMBIC_CONFIG_FILE)
    revision = TAGGED_REVISIONS.get(revision, revision)

    if revision and current_revision:
        script = ScriptDirectory.from_config(Config(ALEMBIC_CONFIG_FILE))

        try:
            script.walk_revisions(base=current_revision, head=revision).__next__()
            action = "upgrade"
        except CommandError:
            action = "downgrade"

    else:
        revision = revision or "head"
        action = "upgrade"

    alembic_args = ["-c", ALEMBIC_CONFIG_FILE, action, revision]
    typer.echo(f"command: alembic {' '.join(alembic_args)}")

    alembic.config.main(argv=alembic_args)


if __name__ == "__main__":
    typer.run(migrate_db)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
from typing import List, Union

from argilla.client.models import TextClassificationRecord
from argilla.training.transformers import ArgillaTransformersTrainer
from argilla.training.utils import get_default_args
from argilla.utils.dependency import require_version


class ArgillaSetFitTrainer(ArgillaTransformersTrainer):
    _logger = logging.getLogger("ArgillaSetFitTrainer")
    _logger.setLevel(logging.INFO)

    require_version("torch")
    require_version("datasets")
    require_version("transformers")
    require_version("setfit>=0.6")

    def __init__(self, *args, **kwargs):
        if kwargs.get("model") is None and "model" in kwargs:
            kwargs["model"] = "all-MiniLM-L6-v2"
        self.multi_target_strategy = None
        self._column_mapping = None
        super().__init__(*args, **kwargs)

        if self._record_class is not TextClassificationRecord:
            raise NotImplementedError("SetFit only supports the `TextClassification` task.")

        if self._multi_label:
            self._column_mapping = {"text": "text", "binarized_label": "label"}
            self.multi_target_strategy = "one-vs-rest"
        else:
            self.multi_target_strategy = None
            self._column_mapping = {"text": "text", "label": "label"}
        self.init_training_args()

    def init_training_args(self) -> None:
        from setfit import SetFitModel, SetFitTrainer

        # SetFit only: we get both the HuggingFace Hub args and the SetFit-specific args
        # We get the default args for `_from_pretrained` first to override the shared args
        # with the HuggingFace Hub specific args
        self.setfit_model_kwargs = get_default_args(SetFitModel._from_pretrained)
        self.setfit_model_kwargs.update(get_default_args(SetFitModel.from_pretrained))

        # Due to an inconsistency between `pretrained_model_name_or_path` with both `model_id` and `revision`
        # we pop both `model_id` and `revision`
        self.setfit_model_kwargs.pop("model_id", None)
        self.setfit_model_kwargs.pop("revision", None)

        self.setfit_model_kwargs["pretrained_model_name_or_path"] = self._model
        self.setfit_model_kwargs["multi_target_strategy"] = self.multi_target_strategy
        self.setfit_model_kwargs["device"] = self.device

        self.setfit_trainer_kwargs = get_default_args(SetFitTrainer.__init__)
        self.setfit_trainer_kwargs["column_mapping"] = self._column_mapping
        self.setfit_trainer_kwargs["train_dataset"] = self._train_dataset
        self.setfit_trainer_kwargs["eval_dataset"] = self._eval_dataset
        self.setfit_trainer_kwargs["seed"] = self._seed

        self._setfit_model = None

    def update_config(
        self,
        **kwargs,
    ) -> None:
        """
        Updates the `setfit_model_kwargs` and `setfit_trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        self.setfit_model_kwargs.update((k, v) for k, v in kwargs.items() if k in self.setfit_model_kwargs)
        self.setfit_trainer_kwargs.update((k, v) for k, v in kwargs.items() if k in self.setfit_trainer_kwargs)

    def __repr__(self):
        formatted_string = []
        arg_dict = {
            "'SetFitModel'": self.setfit_model_kwargs,
            "'SetFitTrainer'": self.setfit_trainer_kwargs,
        }
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\n".join(formatted_string)

    def train(self, output_dir: str = None):
        """
        We create a SetFitModel object from a pretrained model, then create a SetFitTrainer object with
        the model, and then train the model
        """
        from setfit import SetFitModel, SetFitTrainer

        self._setfit_model = SetFitModel.from_pretrained(**self.setfit_model_kwargs)
        self.setfit_trainer_kwargs["model"] = self._setfit_model
        self.__trainer = SetFitTrainer(**self.setfit_trainer_kwargs)
        self.__trainer.train()
        if self._eval_dataset:
            self._metrics = self.__trainer.evaluate()
            self._logger.info(self._metrics)
        else:
            self._metrics = None

        if output_dir is not None:
            self.save(output_dir)

    def init_model(self):
        from setfit import SetFitModel

        self._setfit_model = SetFitModel.from_pretrained(**self.setfit_model_kwargs)

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        if self._setfit_model is None:
            self._logger.warning("Using model without fine-tuning.")
            self.init_model()

        str_input = False
        if isinstance(text, str):
            text = [text]
            str_input = True

        predictions = self._setfit_model.predict_proba(text, **kwargs)

        formatted_prediction = []
        for val, pred in zip(text, predictions):
            pred = {self._id2label[idx]: float(p) for idx, p in enumerate(pred)}
            if as_argilla_records:
                pred = self._record_class(
                    text=val, prediction=[(k, v) for k, v in pred.items()], multi_label=self._multi_label
                )
            formatted_prediction.append(pred)

        if str_input:
            formatted_prediction = formatted_prediction[0]
        return formatted_prediction

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified, and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          path (str): the path to save the model to
        """
        if not isinstance(output_dir, str):
            output_dir = str(output_dir)
        self._setfit_model.save_pretrained(output_dir)

        # store dict as json
        with open(output_dir + "/label2id.json", "w") as f:
            json.dump(self._label2id, f)
        with open(output_dir + "/id2label.json", "w") as f:
            json.dump(self._id2label, f)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import tempfile
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Literal, Optional, Union
from uuid import UUID

from pydantic import (
    ValidationError,
    parse_obj_as,
)
from tqdm import tqdm

import argilla as rg
from argilla.client.feedback.config import FeedbackDatasetConfig
from argilla.client.feedback.constants import (
    FETCHING_BATCH_SIZE,
    FIELD_TYPE_TO_PYTHON_TYPE,
    PUSHING_BATCH_SIZE,
)
from argilla.client.feedback.schemas import (
    FeedbackRecord,
    FieldSchema,
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextField,
    TextQuestion,
)
from argilla.client.feedback.training.schemas import (
    TrainingTaskMappingForTextClassification,
)
from argilla.client.feedback.typing import AllowedFieldTypes, AllowedQuestionTypes
from argilla.client.feedback.unification import (
    LabelQuestionStrategy,
    MultiLabelQuestionStrategy,
    RatingQuestionStrategy,
)
from argilla.client.feedback.utils import (
    feedback_dataset_in_argilla,
    generate_pydantic_schema,
)
from argilla.client.models import Framework
from argilla.client.sdk.v1.datasets import api as datasets_api_v1
from argilla.utils.dependency import require_version, requires_version

if TYPE_CHECKING:
    import httpx
    from datasets import Dataset

    from argilla.client.client import Argilla as ArgillaClient
    from argilla.client.sdk.v1.datasets.models import (
        FeedbackDatasetModel,
        FeedbackFieldModel,
        FeedbackQuestionModel,
    )

_LOGGER = logging.getLogger(__name__)


class FeedbackDataset:
    """Class to work with `FeedbackDataset`s either locally, or remotely (Argilla or HuggingFace Hub).

    Args:
        guidelines: contains the guidelines for annotating the dataset.
        fields: contains the fields that will define the schema of the records in the dataset.
        questions: contains the questions that will be used to annotate the dataset.

    Attributes:
        guidelines: contains the guidelines for annotating the dataset.
        fields: contains the fields that will define the schema of the records in the dataset.
        questions: contains the questions that will be used to annotate the dataset.
        records: contains the records of the dataset if any. Otherwise it is an empty list.
        argilla_id: contains the id of the dataset in Argilla, if it has been uploaded (via `self.push_to_argilla()`). Otherwise, it is `None`.

    Raises:
        TypeError: if `guidelines` is not a string.
        TypeError: if `fields` is not a list of `FieldSchema`.
        ValueError: if `fields` does not contain at least one required field.
        TypeError: if `questions` is not a list of `TextQuestion`, `RatingQuestion`,
            `LabelQuestion`, and/or `MultiLabelQuestion`.
        ValueError: if `questions` does not contain at least one required question.

    Examples:
        >>> import argilla as rg
        >>> rg.init(api_url="...", api_key="...")
        >>> dataset = rg.FeedbackDataset(
        ...     fields=[
        ...         rg.TextField(name="text", required=True),
        ...         rg.TextField(name="label", required=True),
        ...     ],
        ...     questions=[
        ...         rg.TextQuestion(
        ...             name="question-1",
        ...             description="This is the first question",
        ...             required=True,
        ...         ),
        ...         rg.RatingQuestion(
        ...             name="question-2",
        ...             description="This is the second question",
        ...             required=True,
        ...             values=[1, 2, 3, 4, 5],
        ...         ),
        ...         rg.LabelQuestion(
        ...             name="question-3",
        ...             description="This is the third question",
        ...             required=True,
        ...             labels=["positive", "negative"],
        ...         ),
        ...         rg.MultiLabelQuestion(
        ...             name="question-4",
        ...             description="This is the fourth question",
        ...             required=True,
        ...             labels=["category-1", "category-2", "category-3"],
        ...         ),
        ...     ],
        ...     guidelines="These are the annotation guidelines.",
        ... )
        >>> dataset.add_records(
        ...     [
        ...         rg.FeedbackRecord(
        ...             fields={"text": "This is the first record", "label": "positive"},
        ...             responses=[{"values": {"question-1": {"value": "This is the first answer"}, "question-2": {"value": 5}}}],
        ...             external_id="entry-1",
        ...         ),
        ...     ]
        ... )
        >>> dataset.records
        [FeedbackRecord(fields={"text": "This is the first record", "label": "positive"}, responses=[ResponseSchema(user_id=None, values={"question-1": ValueSchema(value="This is the first answer"), "question-2": ValueSchema(value=5), "question-3": ValueSchema(value="positive"), "question-4": ValueSchema(value=["category-1"])})], external_id="entry-1")]
        >>> dataset.push_to_argilla(name="my-dataset", workspace="my-workspace")
        >>> dataset.argilla_id
        "..."
        >>> dataset = rg.FeedbackDataset.from_argilla(argilla_id="...")
        >>> dataset.records
        [FeedbackRecord(fields={"text": "This is the first record", "label": "positive"}, responses=[ResponseSchema(user_id=None, values={"question-1": ValueSchema(value="This is the first answer"), "question-2": ValueSchema(value=5), "question-3": ValueSchema(value="positive"), "question-4": ValueSchema(value=["category-1"])})], external_id="entry-1")]
    """

    argilla_id: Optional[UUID] = None

    def __init__(
        self,
        *,
        fields: List[AllowedFieldTypes],
        questions: List[AllowedQuestionTypes],
        guidelines: Optional[str] = None,
    ) -> None:
        """Initializes a `FeedbackDataset` instance locally.

        Args:
            fields: contains the fields that will define the schema of the records in the dataset.
            questions: contains the questions that will be used to annotate the dataset.
            guidelines: contains the guidelines for annotating the dataset. Defaults to `None`.

        Raises:
            TypeError: if `fields` is not a list of `FieldSchema`.
            ValueError: if `fields` does not contain at least one required field.
            TypeError: if `questions` is not a list of `TextQuestion`, `RatingQuestion`,
                `LabelQuestion`, and/or `MultiLabelQuestion`.
            ValueError: if `questions` does not contain at least one required question.
            TypeError: if `guidelines` is not None and not a string.
            ValueError: if `guidelines` is an empty string.

        Examples:
            >>> import argilla as rg
            >>> rg.init(api_url="...", api_key="...")
            >>> dataset = rg.FeedbackDataset(
            ...     fields=[
            ...         rg.TextField(name="text", required=True),
            ...         rg.TextField(name="label", required=True),
            ...     ],
            ...     questions=[
            ...         rg.TextQuestion(
            ...             name="question-1",
            ...             description="This is the first question",
            ...             required=True,
            ...         ),
            ...         rg.RatingQuestion(
            ...             name="question-2",
            ...             description="This is the second question",
            ...             required=True,
            ...             values=[1, 2, 3, 4, 5],
            ...         ),
            ...         rg.LabelQuestion(
            ...             name="question-3",
            ...             description="This is the third question",
            ...             required=True,
            ...             labels=["positive", "negative"],
            ...         ),
            ...         rg.MultiLabelQuestion(
            ...             name="question-4",
            ...             description="This is the fourth question",
            ...             required=True,
            ...             labels=["category-1", "category-2", "category-3"],
            ...         ),
            ...     ],
            ...     guidelines="These are the annotation guidelines.",
            ... )
        """
        if not isinstance(fields, list):
            raise TypeError(f"Expected `fields` to be a list, got {type(fields)} instead.")

        any_required = False
        unique_names = set()
        for field in fields:
            if not isinstance(field, FieldSchema):
                raise TypeError(f"Expected `fields` to be a list of `FieldSchema`, got {type(field)} instead.")
            if field.name in unique_names:
                raise ValueError(f"Expected `fields` to have unique names, got {field.name} twice instead.")
            unique_names.add(field.name)
            if not any_required and field.required:
                any_required = True
        if not any_required:
            raise ValueError("At least one `FieldSchema` in `fields` must be required (`required=True`).")
        self.__fields = fields
        self.__fields_schema = None

        if not isinstance(questions, list):
            raise TypeError(f"Expected `questions` to be a list, got {type(questions)} instead.")

        any_required = False
        unique_names = set()
        for question in questions:
            if not isinstance(question, AllowedQuestionTypes.__args__):
                raise TypeError(
                    "Expected `questions` to be a list of"
                    f" `{'`, `'.join([arg.__name__ for arg in AllowedQuestionTypes.__args__])}` got a"
                    f" question in the list with type {type(question)} instead."
                )
            if question.name in unique_names:
                raise ValueError(f"Expected `questions` to have unique names, got {question.name} twice instead.")
            unique_names.add(question.name)
            if not any_required and question.required:
                any_required = True
        if not any_required:
            raise ValueError("At least one question in `questions` must be required (`required=True`).")
        self.__questions = questions

        if guidelines is not None:
            if not isinstance(guidelines, str):
                raise TypeError(
                    f"Expected `guidelines` to be either None (default) or a string, got {type(guidelines)} instead."
                )
            if len(guidelines) < 1:
                raise ValueError(
                    "Expected `guidelines` to be either None (default) or a non-empty string, minimum length is 1."
                )
        self.__guidelines = guidelines

        self.__records = []
        self.__new_records = []

    def __len__(self) -> int:
        """Returns the number of records in the dataset."""
        return len(self.records)

    def __getitem__(self, key: Union[slice, int]) -> Union[FeedbackRecord, List[FeedbackRecord]]:
        """Returns the record(s) at the given index(es).

        Args:
            key: the index(es) of the record(s) to return. Can either be a single index or a slice.

        Returns:
            Either the record of the given index, or a list with the records at the given indexes.
        """
        if len(self.records) < 1:
            raise RuntimeError(
                "In order to get items from `rg.FeedbackDataset` you need to either add"
                " them first with `add_records` or fetch them from Argilla or"
                " HuggingFace with `fetch_records`."
            )
        if isinstance(key, int) and len(self.records) < key:
            raise IndexError(f"This dataset contains {len(self)} records, so index {key} is out of range.")
        return self.records[key]

    def __del__(self) -> None:
        """When the dataset object is deleted, delete all the records as well to avoid memory leaks."""
        if hasattr(self, "__records") and self.__records is not None:
            del self.__records
        if hasattr(self, "__new_records") and self.__new_records is not None:
            del self.__new_records

    def __enter__(self) -> "FeedbackDataset":
        """Allows the dataset to be used as a context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """When the context manager is exited, delete all the records as well to avoid memory leaks."""
        self.__del__()

    @property
    def guidelines(self) -> str:
        """Returns the guidelines for annotating the dataset."""
        return self.__guidelines

    @guidelines.setter
    def guidelines(self, guidelines: str) -> None:
        """Sets the guidelines for annotating the dataset."""
        if not isinstance(guidelines, str):
            raise TypeError(f"Expected `guidelines` to be a string, got {type(guidelines)} instead.")
        self.__guidelines = guidelines

    @property
    def fields(self) -> List["FeedbackFieldModel"]:
        """Returns the fields that define the schema of the records in the dataset."""
        return self.__fields

    def field_by_name(self, name: str) -> Dict[str, FieldSchema]:
        for item in self.fields:
            if item.name == name:
                return item
        raise KeyError(f"Field with name '{name}' not found in self.fields")

    @property
    def questions(self) -> List["FeedbackQuestionModel"]:
        """Returns the questions that will be used to annotate the dataset."""
        return self.__questions

    def question_by_name(self, name: str) -> Dict[str, "FeedbackQuestionModel"]:
        for item in self.questions:
            if item.name == name:
                return item
        raise KeyError(f"Question with name '{name}' not found in self.questions")

    @property
    def records(self) -> List[FeedbackRecord]:
        """Returns the all the records in the dataset."""
        return self.__records + self.__new_records

    def fetch_records(self) -> None:
        """Fetches the records from Argilla or HuggingFace and stores them locally.

        If the dataset has not been saved in Argilla or HuggingFace, a warning will be
        raised and the current records will be returned instead.
        """
        if not self.argilla_id:
            _LOGGER.warning(
                "No records have been logged into neither Argilla nor HuggingFace, so"
                " no records will be fetched. The current records will be returned"
                " instead."
            )
            return self.records

        # TODO(alvarobartt): create `ArgillaMixIn` and `HuggingFaceMixIn` classes to inherit their specific methods
        if self.argilla_id:
            httpx_client: "httpx.Client" = rg.active_client().http_client.httpx
            first_batch = datasets_api_v1.get_records(
                client=httpx_client, id=self.argilla_id, offset=0, limit=FETCHING_BATCH_SIZE
            ).parsed
            self.__records = parse_obj_as(List[FeedbackRecord], first_batch.items)
            current_batch = 1
            # TODO(alvarobartt): use `total` from Argilla Metrics API
            with tqdm(
                initial=current_batch,
                desc="Fetching records from Argilla",
            ) as pbar:
                while True:
                    batch = datasets_api_v1.get_records(
                        client=httpx_client,
                        id=self.argilla_id,
                        offset=FETCHING_BATCH_SIZE * current_batch,
                        limit=FETCHING_BATCH_SIZE,
                    ).parsed
                    records = parse_obj_as(List[FeedbackRecord], batch.items)
                    self.__records += records
                    current_batch += 1
                    pbar.update(1)

                    if len(records) < FETCHING_BATCH_SIZE:
                        break

    def add_records(
        self,
        records: Union[FeedbackRecord, Dict[str, Any], List[Union[FeedbackRecord, Dict[str, Any]]]],
    ) -> None:
        """Adds the given records to the dataset, and stores them locally. If you're planning to add those
        records either to Argilla or HuggingFace, make sure to call `push_to_argilla` or `push_to_huggingface`,
        respectively, after adding the records.

        Args:
            records: the records to add to the dataset. Can be a single record, a list of records or a dictionary
                with the fields of the record.

        Raises:
            ValueError: if the given records are an empty list.
            ValueError: if the given records are neither: `FeedbackRecord`, list of `FeedbackRecord`,
                list of dictionaries as a record or dictionary as a record.
            ValueError: if the given records do not match the expected schema.

        Examples:
            >>> import argilla as rg
            >>> rg.init(api_url="...", api_key="...")
            >>> dataset = rg.FeedbackDataset(
            ...     fields=[
            ...         rg.TextField(name="text", required=True),
            ...         rg.TextField(name="label", required=True),
            ...     ],
            ...     questions=[
            ...         rg.TextQuestion(
            ...             name="question-1",
            ...             description="This is the first question",
            ...             required=True,
            ...         ),
            ...         rg.RatingQuestion(
            ...             name="question-2",
            ...             description="This is the second question",
            ...             required=True,
            ...             values=[1, 2, 3, 4, 5],
            ...         ),
            ...         rg.LabelQuestion(
            ...             name="question-3",
            ...             description="This is the third question",
            ...             required=True,
            ...             labels=["positive", "negative"],
            ...         ),
            ...         rg.MultiLabelQuestion(
            ...             name="question-4",
            ...             description="This is the fourth question",
            ...             required=True,
            ...             labels=["category-1", "category-2", "category-3"],
            ...         ),
            ...     ],
            ...     guidelines="These are the annotation guidelines.",
            ... )
            >>> dataset.add_records(
            ...     [
            ...         rg.FeedbackRecord(
            ...             fields={"text": "This is the first record", "label": "positive"},
            ...             responses=[{"values": {"question-1": {"value": "This is the first answer"}, "question-2": {"value": 5}, "question-3": {"value": "positive"}, "question-4": {"value": ["category-1"]}}}],
            ...             external_id="entry-1",
            ...         ),
            ...     ]
            ... )
            >>> dataset.records
            [FeedbackRecord(fields={"text": "This is the first record", "label": "positive"}, responses=[ResponseSchema(user_id=None, values={"question-1": ValueSchema(value="This is the first answer"), "question-2": ValueSchema(value=5), "question-3": ValueSchema(value="positive"), "question-4": ValueSchema(value=["category-1"])})], external_id="entry-1")]
        """
        if isinstance(records, list):
            if len(records) == 0:
                raise ValueError("Expected `records` to be a non-empty list of `dict` or `rg.FeedbackRecord`.")
            new_records = []
            for record in records:
                if isinstance(record, dict):
                    new_records.append(FeedbackRecord(**record))
                elif isinstance(record, FeedbackRecord):
                    new_records.append(record)
                else:
                    raise ValueError(
                        f"Expected `records` to be a list of `dict` or `rg.FeedbackRecord`, got type {type(record)} instead."
                    )
            records = new_records
        elif isinstance(records, dict):
            records = [FeedbackRecord(**records)]
        elif isinstance(records, FeedbackRecord):
            records = [records]
        else:
            raise ValueError(
                f"Expected `records` to be a `dict` or `rg.FeedbackRecord`, got type {type(records)} instead."
            )

        if self.__fields_schema is None:
            self.__fields_schema = generate_pydantic_schema(self.fields)

        for record in records:
            try:
                self.__fields_schema.parse_obj(record.fields)
            except ValidationError as e:
                raise ValueError(
                    f"`rg.FeedbackRecord.fields` does not match the expected schema, with exception: {e}"
                ) from e

        if len(self.__new_records) > 0:
            self.__new_records += records
        else:
            self.__new_records = records

    def iter(self, batch_size: Optional[int] = FETCHING_BATCH_SIZE) -> Iterator[List[FeedbackRecord]]:
        """Returns an iterator over the records in the dataset.

        Args:
            batch_size: the size of the batches to return. Defaults to 100.
        """
        for i in range(0, len(self.records), batch_size):
            yield self.records[i : i + batch_size]

    def push_to_argilla(
        self,
        name: Optional[str] = None,
        workspace: Optional[Union[str, rg.Workspace]] = None,
        show_progress: bool = False,
    ) -> None:
        """Pushes the `FeedbackDataset` to Argilla. If the dataset has been previously pushed to Argilla, it will be updated
        with the new records.

        Note that you may need to `rg.init(...)` with your Argilla credentials before calling this function, otherwise
        the default http://localhost:6900 will be used, which will fail if Argilla is not deployed locally.

        Args:
            name: the name of the dataset to push to Argilla. If not provided, the `argilla_id` will be used if the dataset
                has been previously pushed to Argilla.
            workspace: the workspace where to push the dataset to. If not provided, the active workspace will be used.
            show_progress: the option to choose to show/hide tqdm progress bar while looping over records.
        """
        client: "ArgillaClient" = rg.active_client()
        httpx_client: "httpx.Client" = client.http_client.httpx

        if name is None:
            if self.argilla_id is None:
                _LOGGER.warning(
                    "No `name` has been provided, and no dataset has been pushed to Argilla yet, so no records will be"
                    " pushed to Argilla."
                )
                return

            if len(self.__new_records) < 1:
                _LOGGER.warning(
                    "No new records have been added to the current `FeedbackTask` dataset, so no records will be pushed"
                    " to Argilla."
                )
                return

            try:
                for i in range(0, len(self.__new_records), PUSHING_BATCH_SIZE):
                    datasets_api_v1.add_records(
                        client=httpx_client,
                        id=self.argilla_id,
                        records=[
                            record.dict()
                            for record in tqdm(
                                self.__new_records[i : i + PUSHING_BATCH_SIZE],
                                desc="Pushing records to Argilla...",
                                disable=show_progress,
                            )
                        ],
                    )
                self.__records += self.__new_records
                self.__new_records = []
            except Exception as e:
                raise Exception(
                    f"Failed while adding new records to the current `FeedbackTask` dataset in Argilla with exception: {e}"
                ) from e
        else:
            if workspace is None:
                workspace = rg.Workspace.from_name(client.get_workspace())

            if isinstance(workspace, str):
                workspace = rg.Workspace.from_name(workspace)

            dataset = feedback_dataset_in_argilla(name=name, workspace=workspace)
            if dataset is not None:
                raise RuntimeError(
                    f"Dataset with name=`{name}` and workspace=`{workspace.name}` already exists in Argilla, please"
                    " choose another name and/or workspace."
                )

            try:
                new_dataset: "FeedbackDatasetModel" = datasets_api_v1.create_dataset(
                    client=httpx_client, name=name, workspace_id=workspace.id, guidelines=self.guidelines
                ).parsed
                argilla_id = new_dataset.id
            except Exception as e:
                raise Exception(
                    f"Failed while creating the `FeedbackTask` dataset in Argilla with exception: {e}"
                ) from e

            def delete_dataset(dataset_id: UUID) -> None:
                try:
                    datasets_api_v1.delete_dataset(client=httpx_client, id=dataset_id)
                except Exception as e:
                    raise Exception(
                        f"Failed while deleting the `FeedbackTask` dataset with ID '{dataset_id}' from Argilla with"
                        f" exception: {e}"
                    ) from e

            for field in self.fields:
                try:
                    datasets_api_v1.add_field(client=httpx_client, id=argilla_id, field=field.dict())
                except Exception as e:
                    delete_dataset(dataset_id=argilla_id)
                    raise Exception(
                        f"Failed while adding the field '{field.name}' to the `FeedbackTask` dataset in Argilla with"
                        f" exception: {e}"
                    ) from e

            for question in self.questions:
                try:
                    datasets_api_v1.add_question(client=httpx_client, id=argilla_id, question=question.dict())
                except Exception as e:
                    delete_dataset(dataset_id=argilla_id)
                    raise Exception(
                        f"Failed while adding the question '{question.name}' to the `FeedbackTask` dataset in Argilla"
                        f" with exception: {e}"
                    ) from e

            try:
                datasets_api_v1.publish_dataset(client=httpx_client, id=argilla_id)
            except Exception as e:
                delete_dataset(dataset_id=argilla_id)
                raise Exception(
                    f"Failed while publishing the `FeedbackTask` dataset in Argilla with exception: {e}"
                ) from e

            for batch in self.iter():
                try:
                    datasets_api_v1.add_records(
                        client=httpx_client,
                        id=argilla_id,
                        records=[
                            record.dict()
                            for record in tqdm(batch, desc="Pushing records to Argilla...", disable=show_progress)
                        ],
                    )
                except Exception as e:
                    delete_dataset(dataset_id=argilla_id)
                    raise Exception(
                        f"Failed while adding the records to the `FeedbackTask` dataset in Argilla with exception: {e}"
                    ) from e

            self.__records += self.__new_records
            self.__new_records = []

            if self.argilla_id is not None:
                _LOGGER.warning(
                    "Since the current object is already a `FeedbackDataset` pushed to Argilla, you'll keep on"
                    " interacting with the same dataset in Argilla, even though the one you just pushed holds a"
                    f" different ID ({argilla_id}). So on, if you want to switch to the newly pushed `FeedbackDataset`"
                    f" instead, please use `FeedbackDataset.from_argilla(id='{argilla_id}')`."
                )
                return
            self.argilla_id = argilla_id

    @classmethod
    def from_argilla(
        cls,
        name: Optional[str] = None,
        *,
        workspace: Optional[str] = None,
        id: Optional[str] = None,
        with_records: bool = True,
    ) -> "FeedbackDataset":
        """Retrieves an existing `FeedbackDataset` from Argilla (must have been pushed in advance).

        Note that even no argument is mandatory, you must provide either the `name`, the combination of
        `name` and `workspace`, or the `id`, otherwise an error will be raised.

        Args:
            name: the name of the `FeedbackDataset` to retrieve from Argilla. Defaults to `None`.
            workspace: the workspace of the `FeedbackDataset` to retrieve from Argilla. If not provided, the active
                workspace will be used.
            id: the ID of the `FeedbackDataset` to retrieve from Argilla. Defaults to `None`.
            with_records: whether to retrieve the records of the `FeedbackDataset` from Argilla. Defaults to `True`.

        Returns:
            The `FeedbackDataset` retrieved from Argilla.

        Raises:
            ValueError: if no `FeedbackDataset` with the provided `name` and `workspace` exists in Argilla.
            ValueError: if no `FeedbackDataset` with the provided `id` exists in Argilla.

        Examples:
            >>> import argilla as rg
            >>> rg.init(...)
            >>> dataset = rg.FeedbackDataset.from_argilla(name="my_dataset")
        """
        httpx_client: "httpx.Client" = rg.active_client().http_client.httpx

        existing_dataset = feedback_dataset_in_argilla(name=name, workspace=workspace, id=id)
        if existing_dataset is None:
            raise ValueError(
                f"Could not find a `FeedbackTask` dataset in Argilla with name='{name}'."
                if name and not workspace
                else (
                    "Could not find a `FeedbackTask` dataset in Argilla with"
                    f" name='{name}' and workspace='{workspace}'."
                    if name and workspace
                    else (f"Could not find a `FeedbackTask` dataset in Argilla with ID='{id}'.")
                )
            )

        fields = []
        for field in datasets_api_v1.get_fields(client=httpx_client, id=existing_dataset.id).parsed:
            base_field = field.dict(include={"name", "title", "required"})
            if field.settings["type"] == "text":
                field = TextField(**base_field, use_markdown=field.settings["use_markdown"])
            else:
                raise ValueError(
                    f"Field '{field.name}' is not a supported field in the current Python package version,"
                    " supported field types are: `TextField`."
                )
            fields.append(field)
        questions = []
        for question in datasets_api_v1.get_questions(client=httpx_client, id=existing_dataset.id).parsed:
            question_dict = question.dict(include={"name", "title", "description", "required"})
            if question.settings["type"] == "rating":
                question = RatingQuestion(**question_dict, values=[v["value"] for v in question.settings["options"]])
            elif question.settings["type"] == "text":
                question = TextQuestion(**question_dict, use_markdown=question.settings["use_markdown"])
            elif question.settings["type"] in ["label_selection", "multi_label_selection", "ranking"]:
                if all([label["value"] == label["text"] for label in question.settings["options"]]):
                    labels = [label["value"] for label in question.settings["options"]]
                else:
                    labels = {label["value"]: label["text"] for label in question.settings["options"]}

                if question.settings["type"] == "label_selection":
                    question = LabelQuestion(
                        **question_dict, labels=labels, visible_labels=question.settings["visible_options"]
                    )
                elif question.settings["type"] == "multi_label_selection":
                    question = MultiLabelQuestion(
                        **question_dict, labels=labels, visible_labels=question.settings["visible_options"]
                    )
                elif question.settings["type"] == "ranking":
                    question = RankingQuestion(**question_dict, values=labels)
            else:
                raise ValueError(
                    f"Question '{question.name}' is not a supported question in the current Python package"
                    f" version, supported question types are: `{'`, `'.join([arg.__name__ for arg in AllowedQuestionTypes.__args__])}`."
                )
            questions.append(question)
        self = cls(
            fields=fields,
            questions=questions,
            guidelines=existing_dataset.guidelines or None,
        )
        self.argilla_id = existing_dataset.id
        if with_records:
            self.fetch_records()
        return self

    @requires_version("datasets")
    def format_as(self, format: Literal["datasets"]) -> "Dataset":
        """Formats the `FeedbackDataset` as a `datasets.Dataset` object.

        Args:
            format: the format to use to format the `FeedbackDataset`. Currently supported formats are:
                `datasets`.

        Returns:
            The `FeedbackDataset.records` formatted as a `datasets.Dataset` object.

        Raises:
            ValueError: if the provided format is not supported.

        Examples:
            >>> import argilla as rg
            >>> rg.init(...)
            >>> dataset = rg.FeedbackDataset.from_argilla(name="my-dataset")
            >>> huggingface_dataset = dataset.format_as("datasets")
        """
        if format == "datasets":
            from datasets import Dataset, Features, Sequence, Value

            dataset = {"metadata": []}
            features = {}
            for field in self.fields:
                if field.settings["type"] not in FIELD_TYPE_TO_PYTHON_TYPE.keys():
                    raise ValueError(
                        f"Field {field.name} has an unsupported type: {field.settings['type']}, for the moment"
                        f" only the following types are supported: {list(FIELD_TYPE_TO_PYTHON_TYPE.keys())}"
                    )
                features[field.name] = Value(dtype="string", id="field")
                if field.name not in dataset:
                    dataset[field.name] = []
            for question in self.questions:
                if question.settings["type"] in ["text", "label_selection"]:
                    value = Value(dtype="string")
                elif question.settings["type"] == "rating":
                    value = Value(dtype="int32")
                elif question.settings["type"] == "ranking":
                    value = Sequence({"rank": Value(dtype="uint8"), "value": Value(dtype="string")})
                elif question.settings["type"] in "multi_label_selection":
                    value = Sequence(Value(dtype="string"))
                else:
                    raise ValueError(
                        f"Question {question.name} has an unsupported type: {question.settings['type']}, for the"
                        " moment only the following types are supported: 'text', 'rating', 'label_selection',"
                        " 'multi_label_selection', and 'ranking'."
                    )
                # TODO(alvarobartt): if we constraint ranges from 0 to N, then we can use `ClassLabel` for ratings
                features[question.name] = Sequence(
                    {
                        "user_id": Value(dtype="string"),
                        "value": value,
                        "status": Value(dtype="string"),
                    },
                    id="question",
                )
                if question.name not in dataset:
                    dataset[question.name] = []
            features["external_id"] = Value(dtype="string", id="external_id")
            dataset["external_id"] = []

            for record in self.records:
                for field in self.fields:
                    dataset[field.name].append(record.fields[field.name])
                for question in self.questions:
                    if not record.responses:
                        dataset[question.name].append(None)
                        continue
                    responses = []
                    for response in record.responses:
                        if question.name not in response.values:
                            responses.append(None)
                            continue
                        if question.settings["type"] == "ranking":
                            responses.append([r.dict() for r in response.values[question.name].value])
                        else:
                            responses.append(response.values[question.name].value)
                    dataset[question.name].append(
                        {
                            "user_id": [r.user_id for r in record.responses],
                            "value": responses,
                            "status": [r.status for r in record.responses],
                        }
                    )
                dataset["metadata"].append(json.dumps(record.metadata) if record.metadata else None)
                dataset["external_id"].append(record.external_id or None)

            if len(dataset["metadata"]) > 0:
                features["metadata"] = Value(dtype="string")
            else:
                del dataset["metadata"]

            return Dataset.from_dict(
                dataset,
                features=Features(features),
            )
        raise ValueError(f"Unsupported format '{format}'.")

    @requires_version("huggingface_hub")
    def push_to_huggingface(self, repo_id: str, generate_card: Optional[bool] = True, *args, **kwargs) -> None:
        """Pushes the `FeedbackDataset` to the HuggingFace Hub. If the dataset has been previously pushed to the
        HuggingFace Hub, it will be updated instead. Note that some params as `private` have no effect at all
        when a dataset is previously uploaded to the HuggingFace Hub.

        Args:
            repo_id: the ID of the HuggingFace Hub repo to push the `FeedbackDataset` to.
            generate_card: whether to generate a dataset card for the `FeedbackDataset` in the HuggingFace Hub. Defaults
                to `True`.
            *args: the args to pass to `datasets.Dataset.push_to_hub`.
            **kwargs: the kwargs to pass to `datasets.Dataset.push_to_hub`.
        """
        import huggingface_hub
        from huggingface_hub import DatasetCardData, HfApi
        from packaging.version import parse as parse_version

        if parse_version(huggingface_hub.__version__) < parse_version("0.14.0"):
            _LOGGER.warning(
                "Recommended `huggingface_hub` version is 0.14.0 or higher, and you have"
                f" {huggingface_hub.__version__}, so in case you have any issue when pushing the dataset to the"
                " HuggingFace Hub upgrade it as `pip install huggingface_hub --upgrade`."
            )

        if len(self) < 1:
            raise ValueError(
                "Cannot push an empty `rg.FeedbackDataset` to the HuggingFace Hub, please make sure to add at"
                " least one record, via the method `add_records`."
            )

        hfds = self.format_as("datasets")
        hfds.push_to_hub(repo_id, *args, **kwargs)

        with tempfile.NamedTemporaryFile(mode="w", suffix=".cfg", delete=False) as f:
            f.write(
                FeedbackDatasetConfig(fields=self.fields, questions=self.questions, guidelines=self.guidelines).json()
            )
            f.flush()

            HfApi().upload_file(
                path_or_fileobj=f.name,
                path_in_repo="argilla.cfg",
                repo_id=repo_id,
                repo_type="dataset",
                token=kwargs.get("token"),
            )

        if generate_card:
            from argilla.client.feedback.card import (
                ArgillaDatasetCard,
                size_categories_parser,
            )

            card = ArgillaDatasetCard.from_template(
                card_data=DatasetCardData(
                    size_categories=size_categories_parser(len(self.records)),
                    tags=["rlfh", "argilla", "human-feedback"],
                ),
                repo_id=repo_id,
                argilla_fields=self.fields,
                argilla_questions=self.questions,
                argilla_guidelines=self.guidelines,
                argilla_record=json.loads(self.records[0].json()),
                huggingface_record=hfds[0],
            )
            card.push_to_hub(repo_id, repo_type="dataset", token=kwargs.get("token"))

    @classmethod
    @requires_version("datasets")
    @requires_version("huggingface_hub")
    def from_huggingface(cls, repo_id: str, *args: Any, **kwargs: Any) -> "FeedbackDataset":
        """Loads a `FeedbackDataset` from the HuggingFace Hub.

        Args:
            repo_id: the ID of the HuggingFace Hub repo to load the `FeedbackDataset` from.
            *args: the args to pass to `datasets.Dataset.load_from_hub`.
            **kwargs: the kwargs to pass to `datasets.Dataset.load_from_hub`.

        Returns:
            A `FeedbackDataset` loaded from the HuggingFace Hub.
        """
        import huggingface_hub
        from datasets import DatasetDict, load_dataset
        from huggingface_hub import hf_hub_download
        from packaging.version import parse as parse_version

        if parse_version(huggingface_hub.__version__) < parse_version("0.14.0"):
            _LOGGER.warning(
                "Recommended `huggingface_hub` version is 0.14.0 or higher, and you have"
                f" {huggingface_hub.__version__}, so in case you have any issue when pushing the dataset to the"
                " HuggingFace Hub upgrade it as `pip install huggingface_hub --upgrade`."
            )

        if "token" in kwargs:
            auth = kwargs.pop("token")
        elif "use_auth_token" in kwargs:
            auth = kwargs.pop("use_auth_token")
        else:
            auth = None

        hub_auth = (
            {"use_auth_token": auth}
            if parse_version(huggingface_hub.__version__) < parse_version("0.11.0")
            else {"token": auth}
        )
        config_path = hf_hub_download(
            repo_id=repo_id,
            filename="argilla.cfg",
            repo_type="dataset",
            **hub_auth,
        )
        with open(config_path, "r") as f:
            config = FeedbackDatasetConfig.parse_raw(f.read())

        cls = cls(
            fields=config.fields,
            questions=config.questions,
            guidelines=config.guidelines,
        )

        hfds = load_dataset(repo_id, use_auth_token=auth, *args, **kwargs)
        if isinstance(hfds, DatasetDict) and "split" not in kwargs:
            if len(hfds.keys()) > 1:
                raise ValueError(
                    "Only one dataset can be loaded at a time, use `split` to select a split, available splits"
                    f" are: {', '.join(hfds.keys())}."
                )
            hfds = hfds[list(hfds.keys())[0]]

        for index in range(len(hfds)):
            responses = {}
            for question in cls.questions:
                if hfds[index][question.name] is None or len(hfds[index][question.name]) < 1:
                    continue
                for user_id, value, status in zip(
                    hfds[index][question.name]["user_id"],
                    hfds[index][question.name]["value"],
                    hfds[index][question.name]["status"],
                ):
                    if user_id not in responses:
                        responses[user_id] = {
                            "user_id": user_id,
                            "status": status,
                            "values": {},
                        }
                    if question.settings["type"] == "ranking":
                        value = [{"rank": r, "value": v} for r, v in zip(value["rank"], value["value"])]
                    responses[user_id]["values"].update({question.name: {"value": value}})

            metadata = None
            if "metadata" in hfds[index] and hfds[index]["metadata"] is not None:
                metadata = json.loads(hfds[index]["metadata"])

            cls.__records.append(
                FeedbackRecord(
                    fields={field.name: hfds[index][field.name] for field in cls.fields},
                    metadata=metadata,
                    responses=list(responses.values()) or None,
                    external_id=hfds[index]["external_id"],
                )
            )
        del hfds
        return cls

    def unify_responses(
        self,
        question: Union[str, LabelQuestion, MultiLabelQuestion, RatingQuestion],
        strategy: Union[str, LabelQuestionStrategy, MultiLabelQuestionStrategy, RatingQuestionStrategy],
    ) -> None:
        """Unify responses for a given question using a given strategy"""
        if isinstance(question, str):
            question = self.question_by_name(question)
        if isinstance(strategy, str):
            strategy = LabelQuestionStrategy(strategy)
        strategy.unify_responses(self.records, question)

    def prepare_for_training(
        self,
        framework: Union[Framework, str],
        task_mapping: TrainingTaskMappingForTextClassification,
        train_size: Optional[float] = 1,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
        fetch_records: bool = True,
        lang: Optional[str] = None,
    ):
        if isinstance(framework, str):
            framework = Framework(framework)

        # validate train and test sizes
        if train_size is None:
            train_size = 1
        if test_size is None:
            test_size = 1 - train_size

        # check if all numbers are larger than 0
        if not [abs(train_size), abs(test_size)] == [train_size, test_size]:
            raise ValueError("`train_size` and `test_size` must be larger than 0.")
        # check if train sizes sum up to 1
        if not (train_size + test_size) == 1:
            raise ValueError("`train_size` and `test_size` must sum to 1.")

        if test_size == 0:
            test_size = None

        if fetch_records:
            self.fetch_records()

        if isinstance(task_mapping, TrainingTaskMappingForTextClassification):
            self.unify_responses(question=task_mapping.label.question, strategy=task_mapping.label.strategy)
        else:
            raise ValueError(f"Training data {type(task_mapping)} is not supported yet")

        data = task_mapping._format_data(self.records)
        if framework in [
            Framework.TRANSFORMERS,
            Framework.SETFIT,
            Framework.SPAN_MARKER,
            Framework.PEFT,
        ]:
            return task_mapping._prepare_for_training_with_transformers(
                data=data, train_size=train_size, seed=seed, framework=framework
            )
        elif framework is Framework.SPACY or framework is Framework.SPACY_TRANSFORMERS:
            require_version("spacy")
            import spacy

            if lang is None:
                _LOGGER.warning("spaCy `lang` is not provided. Using `en`(English) as default language.")
                lang = spacy.blank("en")
            elif lang.isinstance(str):
                if len(lang) == 2:
                    lang = spacy.blank(lang)
                else:
                    lang = spacy.load(lang)
            return task_mapping._prepare_for_training_with_spacy(data=data, train_size=train_size, seed=seed, lang=lang)
        elif framework is Framework.SPARK_NLP:
            return task_mapping._prepare_for_training_with_spark_nlp(data=data, train_size=train_size, seed=seed)
        elif framework is Framework.OPENAI:
            return task_mapping._prepare_for_training_with_openai(data=data, train_size=train_size, seed=seed)
        else:
            raise NotImplementedError(
                f"Framework {framework} is not supported. Choose from: {[e.value for e in Framework]}"
            )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

from argilla._constants import DEFAULT_API_KEY, DEFAULT_PASSWORD, DEFAULT_USERNAME
from argilla.server.contexts import accounts
from argilla.server.database import SessionLocal
from argilla.server.models import User, UserRole, Workspace


def create_default(
    api_key: str = typer.Option(default=DEFAULT_API_KEY, help="API key for the user."),
    password: str = typer.Option(default=DEFAULT_PASSWORD, help="Password for the user."),
    quiet: bool = typer.Option(is_flag=True, default=False, help="Run without output."),
):
    """Creates a user with default credentials on database suitable to start experimenting with argilla."""
    with SessionLocal() as session:
        if accounts.get_user_by_username(session, DEFAULT_USERNAME):
            if not quiet:
                typer.echo("User with default username already found on database, will not do anything.")

            return

        session.add(
            User(
                first_name="",
                username=DEFAULT_USERNAME,
                role=UserRole.owner,
                api_key=api_key,
                password_hash=accounts.hash_password(password),
                workspaces=[Workspace(name=DEFAULT_USERNAME)],
            )
        )
        session.commit()

        if not quiet:
            typer.echo("User with default credentials successfully created:")
            typer.echo(f"• username: {DEFAULT_USERNAME!r}")
            typer.echo(f"• password: {password!r}")
            typer.echo(f"• role:     {UserRole.owner.value!r}")
            typer.echo(f"• api_key:  {api_key!r}")


if __name__ == "__main__":
    typer.run(create_default)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Any, Dict, Optional, Tuple

from argilla import TokenClassificationRecord
from argilla.client.api import Api
from argilla.monitoring.base import BaseMonitor
from argilla.monitoring.types import MissingType

try:
    # Conditional modules
    from spacy.language import Language
    from spacy.tokens import Doc
except ModuleNotFoundError:
    Language = MissingType
    Doc = MissingType


class SpacyNERMonitor(BaseMonitor):
    """A spaCy Language wrapper for NLP NER monitoring in argilla"""

    @staticmethod
    def doc2token_classification(
        doc: Doc,
        agent: str,
        metadata: Optional[Dict[str, Any]],
    ) -> TokenClassificationRecord:
        """
        Converts a spaCy `Doc` into a token classification record

        Parameters
        ----------
        doc:
            The spacy doc
        agent:
            Agent to use for the prediction_agent field. Could be the model path or model lang + model version
        metadata:
            Passed on to the `argilla.TokenClassificationRecord`.

        """
        entities = [(ent.label_, ent.start_char, ent.end_char) for ent in doc.ents]
        return TokenClassificationRecord(
            text=doc.text,
            tokens=[t.text for t in doc],
            metadata=metadata or {},
            prediction_agent=agent,
            prediction=entities,
            event_timestamp=datetime.utcnow(),
        )

    def _prepare_log_data(self, docs_info: Tuple[Doc, Optional[Dict[str, Any]]]) -> Dict[str, Any]:
        return dict(
            records=[
                self.doc2token_classification(doc, agent=self.__wrapped__.path.name, metadata=metadata)
                for doc, metadata in docs_info
            ],
            name=self.dataset,
            tags={k: v for k, v in self.__model__.meta.items() if isinstance(v, str)},
            metadata=self.__model__.meta,
        )

    def pipe(self, *args, **kwargs):
        as_tuples = kwargs.get("as_tuples")
        results = self.__model__.pipe(*args, **kwargs)

        log_info = []
        for r in results:
            metadata = {}
            if as_tuples:
                doc, metadata = r  # context
            else:
                doc = r
            if self.is_record_accepted():
                log_info.append((doc, metadata))
            yield r

        self.send_records(log_info)

    def __call__(self, *args, **kwargs):
        metadata = kwargs.pop("metadata", None)
        doc = self.__wrapped__(*args, **kwargs)
        try:
            if self.is_record_accepted():
                self.send_records([(doc, metadata)])
        finally:
            return doc


def ner_monitor(
    nlp: Language,
    api: Api,
    dataset: str,
    sample_rate: float,
    log_interval: float,
) -> Language:
    return SpacyNERMonitor(
        nlp,
        api=api,
        dataset=dataset,
        sample_rate=sample_rate,
        log_interval=log_interval,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.orm import Session

from argilla.server.contexts import datasets
from argilla.server.database import get_db
from argilla.server.models import User
from argilla.server.policies import RecordPolicyV1, authorize
from argilla.server.schemas.v1.records import Response, ResponseCreate
from argilla.server.search_engine import SearchEngine, get_search_engine
from argilla.server.security import auth

router = APIRouter(tags=["records"])


@router.post("/records/{record_id}/responses", status_code=status.HTTP_201_CREATED, response_model=Response)
async def create_record_response(
    *,
    db: Session = Depends(get_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    response_create: ResponseCreate,
    current_user: User = Security(auth.get_current_user),
):
    record = datasets.get_record_by_id(db, record_id)
    if not record:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Record with id `{record_id}` not found",
        )

    authorize(current_user, RecordPolicyV1.create_response(record))

    if datasets.get_response_by_record_id_and_user_id(db, record_id, current_user.id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Response already exists for record with id `{record_id}` and by user with id `{current_user.id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return await datasets.create_response(db, search_engine, record, current_user, response_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, List, Optional

from pydantic import BaseModel, Field, validator

from argilla.server.apis.v0.models.commons.model import (
    BaseAnnotation,
    BaseRecord,
    BaseRecordInputs,
    BaseSearchResults,
    ScoreRange,
    SortableField,
)
from argilla.server.commons.models import PredictionStatus
from argilla.server.schemas.datasets import UpdateDatasetRequest
from argilla.server.services.metrics.models import CommonTasksMetrics
from argilla.server.services.search.model import (
    ServiceBaseRecordsQuery,
    ServiceBaseSearchResultsAggregations,
)


class Text2TextPrediction(BaseModel):
    text: str
    score: float = Field(default=1.0, ge=0.0, le=1.0)


class Text2TextAnnotation(BaseAnnotation):
    @validator("sentences")
    def sort_sentences_by_score(cls, sentences: List[Text2TextPrediction]):
        """Sort provided sentences by score desc"""
        return sorted(sentences, key=lambda x: x.score, reverse=True)

    sentences: List[Text2TextPrediction]


class Text2TextRecordInputs(BaseRecordInputs[Text2TextAnnotation]):
    text: str


class Text2TextRecord(Text2TextRecordInputs, BaseRecord[Text2TextAnnotation]):
    pass


class Text2TextBulkRequest(UpdateDatasetRequest):
    records: List[Text2TextRecordInputs]


class Text2TextQuery(ServiceBaseRecordsQuery):
    score: Optional[ScoreRange] = Field(default=None)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)


class Text2TextSearchAggregations(ServiceBaseSearchResultsAggregations):
    predicted_text: Dict[str, int] = Field(default_factory=dict)
    annotated_text: Dict[str, int] = Field(default_factory=dict)


class Text2TextSearchResults(BaseSearchResults[Text2TextRecord, Text2TextSearchAggregations]):
    pass


class Text2TextMetrics(CommonTasksMetrics[Text2TextRecord]):
    pass


class Text2TextSearchRequest(BaseModel):
    query: Text2TextQuery = Field(default_factory=Text2TextQuery)
    sort: List[SortableField] = Field(default_factory=list)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from typing import Tuple, Type

import httpx
from opensearchpy import OpenSearch
from packaging.version import parse

from argilla.server.daos.backend.base import GenericSearchError
from argilla.server.daos.backend.client_adapters import (
    ElasticsearchClient,
    IClientAdapter,
    OpenSearchClient,
)

_LOGGER = logging.getLogger("argilla")


class ClientAdapterFactory:
    @classmethod
    def get(
        cls,
        hosts: str,
        index_shards: int,
        ssl_verify: bool,
        ca_path: str,
        retry_on_timeout: bool = True,
        max_retries: int = 5,
    ) -> IClientAdapter:
        client_config = dict(
            hosts=hosts,
            verify_certs=ssl_verify,
            ca_certs=ca_path,
            retry_on_timeout=retry_on_timeout,
            max_retries=max_retries,
        )

        version, distribution = cls._fetch_cluster_version_info(client_config)

        (client_class, support_vector_search) = cls._resolve_client_class_with_vector_support(version, distribution)

        return client_class(
            index_shards=index_shards, vector_search_supported=support_vector_search, config_backend=client_config
        )

    @classmethod
    def _resolve_client_class_with_vector_support(cls, version: str, distribution: str) -> Tuple[Type, bool]:
        support_vector_search = True

        if distribution == "elasticsearch" and parse("8.5") <= parse(version):
            if parse("8.5") <= parse(ElasticsearchClient.ES_CLIENT_VERSION):
                client_class = ElasticsearchClient
            else:
                _LOGGER.warning(
                    "Elasticsearch 8.5 backend found but installed\n"
                    "client does not support vectors. Please upgrade your elasticsearch client\n"
                    "if you want to support similarity search in argilla."
                )
                client_class = OpenSearchClient
                support_vector_search = False
        elif distribution == "opensearch" and parse("2.2") <= parse(version):
            client_class = OpenSearchClient
        else:
            client_class = OpenSearchClient
            support_vector_search = False

        return client_class, support_vector_search

    @classmethod
    def _fetch_cluster_version_info(cls, client_config: dict) -> Tuple[str, str]:
        try:
            # All security config will be used here.
            # See here https://opensearch.org/docs/latest/clients/index/#legacy-clients
            client = OpenSearch(**client_config)
            data = client.info()

            version_info = data["version"]
            version: str = version_info["number"]
            distribution: str = version_info.get("distribution", "elasticsearch")

            return version, distribution
        except Exception as error:
            raise GenericSearchError(error)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, validator

from argilla._constants import DEFAULT_MAX_KEYWORD_LENGTH
from argilla.client.models import (
    TokenClassificationRecord as ClientTokenClassificationRecord,
)
from argilla.client.sdk.commons.models import (
    MACHINE_NAME,
    BaseAnnotation,
    BaseRecord,
    PredictionStatus,
    ScoreRange,
    SdkVectorSearch,
    TaskStatus,
    UpdateDatasetRequest,
)


class EntitySpan(BaseModel):
    start: int
    end: int
    label: str = Field(min_length=1, max_length=DEFAULT_MAX_KEYWORD_LENGTH)
    score: float = Field(default=1.0, ge=0.0, le=1.0)


class TokenClassificationAnnotation(BaseAnnotation):
    entities: List[EntitySpan] = Field(default_factory=list)
    score: Optional[float] = None


class CreationTokenClassificationRecord(BaseRecord[TokenClassificationAnnotation]):
    tokens: List[str] = Field(min_items=1)

    text: str

    @validator("text")
    def check_text_content(cls, text: str):
        assert text and text.strip(), "No text or empty text provided"
        return text

    @classmethod
    def from_client(cls, record: ClientTokenClassificationRecord):
        prediction = None
        if record.prediction is not None:
            prediction = TokenClassificationAnnotation(
                entities=[
                    EntitySpan(label=ent[0], start=ent[1], end=ent[2])
                    if len(ent) == 3
                    else EntitySpan(label=ent[0], start=ent[1], end=ent[2], score=ent[3])
                    for ent in record.prediction
                ],
                agent=record.prediction_agent or MACHINE_NAME,
            )

        annotation = None
        if record.annotation is not None:
            annotation = TokenClassificationAnnotation(
                entities=[EntitySpan(label=ent[0], start=ent[1], end=ent[2]) for ent in record.annotation],
                agent=record.annotation_agent or MACHINE_NAME,
            )

        return cls(
            tokens=record.tokens,
            text=record.text,
            prediction=prediction,
            annotation=annotation,
            vectors=cls._from_client_vectors(record.vectors),
            status=record.status,
            id=record.id,
            metadata=record.metadata,
            event_timestamp=record.event_timestamp,
        )


class TokenClassificationRecord(CreationTokenClassificationRecord):
    last_updated: datetime = None
    _predicted: Optional[PredictionStatus] = Field(alias="predicted")

    def to_client(self) -> ClientTokenClassificationRecord:
        return ClientTokenClassificationRecord(
            text=self.text,
            tokens=self.tokens,
            prediction=[(ent.label, ent.start, ent.end, ent.score) for ent in self.prediction.entities]
            if self.prediction
            else None,
            prediction_agent=self.prediction.agent if self.prediction else None,
            annotation=[(ent.label, ent.start, ent.end) for ent in self.annotation.entities]
            if self.annotation
            else None,
            annotation_agent=self.annotation.agent if self.annotation else None,
            vectors=self._to_client_vectors(self.vectors),
            id=self.id,
            event_timestamp=self.event_timestamp,
            status=self.status,
            metadata=self.metadata or {},
            metrics=self.metrics or None,
            search_keywords=self.search_keywords or None,
        )


class TokenClassificationBulkData(UpdateDatasetRequest):
    records: List[CreationTokenClassificationRecord]


class TokenClassificationQuery(BaseModel):
    ids: Optional[List[Union[str, int]]]

    query_text: str = Field(default=None)
    advanced_query_dsl: bool = False
    metadata: Optional[Dict[str, Union[str, List[str]]]] = None

    predicted_as: List[str] = Field(default_factory=list)
    annotated_as: List[str] = Field(default_factory=list)
    annotated_by: List[str] = Field(default_factory=list)
    predicted_by: List[str] = Field(default_factory=list)
    score: Optional[ScoreRange] = Field(default=None)
    status: List[TaskStatus] = Field(default_factory=list)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)

    vector: Optional[SdkVectorSearch] = Field(default=None)
